from subprocess import check_output, check_call, CalledProcessError
from os import unlink, devnull
from os.path import isfile
from tempfile import NamedTemporaryFile
from fractions import Fraction
from six.moves import zip

from . import cglpk
from .wrappers import *

# detect paths to system calls for esolver and gzip
with open(devnull, "w") as DEVNULL:
    try:
        ESOLVER_COMMAND = check_output(["which", "esolver"],
                                       stderr=DEVNULL).strip()
        __esolver_version__ = check_output(["esolver", "-v"], stderr=DEVNULL)
    except CalledProcessError:
        raise RuntimeError("esolver command not found")
    try:
        GZIP_COMMAND = check_output(["which", "gzip"], stderr=DEVNULL).strip()
    except CalledProcessError:
        raise RuntimeError("gzip command not found")
del DEVNULL

solver_name = "esolver"


class Esolver(cglpk.GLP):
    """contain an LP which will be solved through the QSopt_ex

    The LP is stored using a GLPK object, and written out to an
    LP file which is then solved by the esolver command."""

    def __init__(self, cobra_model=None):
        cglpk.GLP.__init__(self, cobra_model)
        self.solution_filepath = None
        self.basis_filepath = None
        self.rational_solution = False
        self.verbose = False
        self.clean_up = True  # clean up files

    def _clean(self, filename):
        """remove old files"""
        if self.clean_up and filename is not None and isfile(filename):
            unlink(filename)

    def set_parameter(self, parameter_name, value):
        if parameter_name == "GLP":
            raise Exception("can not be set this way")
        if parameter_name == "objective_sense":
            self.set_objective_sense(value)
        if not hasattr(self, parameter_name):
            raise ValueError("Unkonwn parameter '%s'" % parameter_name)
        setattr(self, parameter_name, value)

    def solve_problem(self, **solver_parameters):
        if "objective_sense" in solver_parameters:
            self.set_objective_sense(solver_parameters.pop("objective_sense"))
        for key, value in solver_parameters.items():
            self.set_parameter(key, value)
        # remove the old solution file
        self._clean(self.solution_filepath)
        with NamedTemporaryFile(suffix=".lp", delete=False) as f:
            lp_filepath = f.name
        self.write(lp_filepath)
        existing_basis = self.basis_filepath
        with NamedTemporaryFile(suffix=".bas", delete=False) as f:
            self.basis_filepath = f.name
        with NamedTemporaryFile(suffix=".sol") as f:
            self.solution_filepath = f.name
        command = [ESOLVER_COMMAND, "-b", self.basis_filepath,
                   "-O", self.solution_filepath[:-4]]
        if existing_basis is not None and isfile(existing_basis):
            command.extend(["-B", existing_basis])
        command.extend(["-L", lp_filepath])
        command_kwargs = {}
        if self.verbose:
            print(" ".join(command))
            DEVNULL = None
        else:
            DEVNULL = open(devnull, 'wb')
            command_kwargs["stdout"] = DEVNULL
            command_kwargs["stderr"] = DEVNULL
        try:
            check_call(command, **command_kwargs)
            failed = False
        except CalledProcessError as e:
            failed = True
        if failed:
            self.basis_filepath = existing_basis
            existing_basis = None
            # Sometimes on failure a solution isn't written out
            if not isfile(self.solution_filepath):
                with open(self.solution_filepath, "w") as outfile:
                    outfile.write("=infeasible\n")
        elif isfile(self.solution_filepath + ".gz"):
            # the solution may be written out compressed
            check_call([GZIP_COMMAND, "-d", self.solution_filepath + ".gz"])
        if DEVNULL is not None:
            DEVNULL.close()
        self._clean(lp_filepath)
        self._clean(existing_basis)  # replaced with the new basis

    def get_status(self):
        with open(self.solution_filepath) as infile:
            return infile.readline().split("=")[1].strip().lower()

    def _format(self, value):
        """convert a string value into either a fraction or float"""
        value = Fraction(value)
        return value if self.rational_solution else float(value)

    def get_objective_value(self):
        with open(self.solution_filepath) as infile:
            status = infile.readline().split("=")[1].strip().lower()
            if status != "optimal":
                raise RuntimeError("status not optimal")
            infile.readline()
            return self._format(infile.readline().split("=")[1].strip())

    def format_solution(self, cobra_model):
        m = cobra_model
        solution = m.solution.__class__(None)
        with open(self.solution_filepath) as infile:
            solution.status = infile.readline().split("=")[1].strip().lower()
            if solution.status != "optimal":
                return solution
            infile.readline()
            solution.f = self._format(Fraction(infile.readline()
                                               .split("=")[1].strip()))
            infile.readline()
            value_dict = {}
            for line in infile:
                if line.endswith(":\n"):
                    break
                varname, value = line.split("=")
                value_dict[varname.strip()] = self._format(value.strip())
            dual_dict = {}
            for line in infile:
                if line.endswith(":\n"):
                    break
                varname, value = line.split("=")
                dual_dict[varname.strip()] = self._format(value.strip())
        solution.x = [value_dict.get("x_%d" % (i + 1), 0)
                      for i in range(len(m.reactions))]
        solution.x_dict = {r.id: v for r, v in zip(m.reactions, solution.x)}
        solution.y = [dual_dict.get("r_%d" % (i + 1), 0)
                      for i in range(len(m.metabolites))]
        solution.y_dict = {m.id: v for m, v in zip(m.metabolites, solution.y)}
        return solution


# wrappers for the classmethods at the module level
create_problem = Esolver.create_problem
solve = Esolver.solve



from warnings import warn

from six import string_types, iteritems

from ..manipulation import delete_model_genes, undelete_model_genes
from ..manipulation.delete import find_gene_knockout_reactions
from ..solvers import solver_dict, get_solver_name
try:
    import scipy
except ImportError:
    moma = None
else:
    from . import moma


def single_deletion(cobra_model, element_list=None,
                    element_type='gene', **kwargs):
    """Wrapper for single_gene_deletion and single_reaction_deletion

    .. deprecated :: 0.4
        Use single_reaction_deletion and single_gene_deletion
    """
    warn("deprecated - use single_reaction_deletion and single_gene_deletion")
    if element_type == "reaction":
        return single_reaction_deletion(cobra_model, element_list, **kwargs)
    elif element_type == "gene":
        return single_gene_deletion(cobra_model, element_list, **kwargs)
    else:
        raise Exception("unknown element type")


def single_reaction_deletion(cobra_model, reaction_list=None, solver=None,
                             method="fba", **solver_args):
    """sequentially knocks out each reaction in a model

    reaction_list: list of reaction_ids or cobra.Reaction

    method: "fba" or "moma"

    returns ({reaction_id: growth_rate}, {reaction_id: status})"""
    if reaction_list is None:
        reaction_list = cobra_model.reactions
    else:
        reaction_list = [cobra_model.reactions.get_by_id(i)
                         if isinstance(i, string_types) else i
                         for i in reaction_list]
    if method == "fba":
        return single_reaction_deletion_fba(cobra_model, reaction_list,
                                            solver=solver, **solver_args)
    elif method == "moma":
        return single_reaction_deletion_moma(cobra_model, reaction_list,
                                             solver=solver, **solver_args)
    else:
        raise ValueError("Unknown deletion method '%s'" % method)


def single_reaction_deletion_fba(cobra_model, reaction_list, solver=None,
                                 **solver_args):
    """sequentially knocks out each reaction in a model using FBA

    reaction_list: list of reaction_ids or cobra.Reaction

    method: "fba" or "moma"

    returns ({reaction_id: growth_rate}, {reaction_id: status})"""

    solver = solver_dict[get_solver_name() if solver is None else solver]
    lp = solver.create_problem(cobra_model)

    growth_rate_dict = {}
    status_dict = {}
    for reaction in reaction_list:
        old_bounds = (reaction.lower_bound, reaction.upper_bound)
        index = cobra_model.reactions.index(reaction)
        solver.change_variable_bounds(lp, index, 0., 0.)
        solver.solve_problem(lp, **solver_args)
        # get the status and growth rate
        status = solver.get_status(lp)
        status_dict[reaction.id] = status
        growth_rate_dict[reaction.id] = solver.get_objective_value(lp) \
            if status == "optimal" else 0.
        # reset the problem
        solver.change_variable_bounds(lp, index, old_bounds[0], old_bounds[1])
    return (growth_rate_dict, status_dict)


def single_reaction_deletion_moma(cobra_model, reaction_list, solver=None,
                                  **solver_args):
    """sequentially knocks out each reaction in a model using MOMA

    reaction_list: list of reaction_ids or cobra.Reaction


    returns ({reaction_id: growth_rate}, {reaction_id: status})"""
    # The same function can not be used because MOMA can not re-use the
    # same LP object. Problem re-use leads to incorrect solutions.
    if moma is None:
        raise RuntimeError("scipy required for moma")
    solver = solver_dict[solver if solver else get_solver_name(qp=True)]
    moma_model, moma_obj = moma.create_euclidian_moma_model(cobra_model)

    growth_rate_dict = {}
    status_dict = {}
    for reaction in reaction_list:
        index = cobra_model.reactions.index(reaction)
        solution = moma.moma_knockout(moma_model, moma_objective, (index,),
                                      solver=solver, **solver_args)
        status_dict[reaction.id] = solution.status
        growth_rate_dict[reaction.id] = solution.f
    return (growth_rate_dict, status_dict)


def single_gene_deletion(cobra_model, gene_list=None, solver=None,
                         method="fba", **solver_args):
    """sequentially knocks out each gene in a model

    gene_list: list of gene_ids or cobra.Gene

    method: "fba" or "moma"

    returns ({gene_id: growth_rate}, {gene_id: status})"""
    if gene_list is None:
        gene_list = cobra_model.genes
    else:
        gene_list = [cobra_model.genes.get_by_id(i)
                     if isinstance(i, string_types) else i for i in gene_list]

    if method == "fba":
        return single_gene_deletion_fba(cobra_model, gene_list,
                                        solver=solver, **solver_args)
    elif method == "moma":
        return single_gene_deletion_moma(cobra_model, gene_list,
                                         solver=solver, **solver_args)
    else:
        raise ValueError("Unknown deletion method '%s'" % method)


def single_gene_deletion_fba(cobra_model, gene_list, solver=None,
                             **solver_args):

    solver = solver_dict[get_solver_name() if solver is None else solver]
    lp = solver.create_problem(cobra_model)

    growth_rate_dict = {}
    status_dict = {}
    for gene in gene_list:
        old_bounds = {}
        for reaction in find_gene_knockout_reactions(cobra_model, [gene]):
            index = cobra_model.reactions.index(reaction)
            old_bounds[index] = (reaction.lower_bound, reaction.upper_bound)
            solver.change_variable_bounds(lp, index, 0., 0.)
        solver.solve_problem(lp, **solver_args)
        # get the status and growth rate
        status = solver.get_status(lp)
        status_dict[gene.id] = status
        growth_rate = solver.get_objective_value(lp) \
            if status == "optimal" else 0.
        growth_rate_dict[gene.id] = growth_rate
        # reset the problem
        for index, bounds in iteritems(old_bounds):
            solver.change_variable_bounds(lp, index, bounds[0], bounds[1])
    return (growth_rate_dict, status_dict)


def single_gene_deletion_moma(cobra_model, gene_list, solver=None,
                              **solver_args):
    if moma is None:
        raise RuntimeError("scipy required for moma")
    solver = solver if solver else get_solver_name(qp=True)
    moma_model, moma_obj = moma.create_euclidian_moma_model(cobra_model)

    growth_rate_dict = {}
    status_dict = {}
    for gene in gene_list:
        delete_model_genes(moma_model, [gene.id])
        solution = moma.solve_moma_model(moma_model, moma_obj,
                                         solver=solver, **solver_args)
        status_dict[gene.id] = solution.status
        growth_rate_dict[gene.id] = solution.f
        undelete_model_genes(moma_model)
    return (growth_rate_dict, status_dict)



"""
Wrappers for solvers with an object oriented interface. This creates
functions to call the objects' functions.

The create_problem and solve functions are not included because they
are classmethods. They should be included by specifying
create_problem = PROBLEM_CLASS.create_problem
where PROBLEM_CLASS is the solver class (i.e. GLP, esolver, etc.)
"""


def set_objective_sense(lp, objective_sense="maximize"):
    return lp.set_objective_sense(lp, objective_sense=objective_sense)


def change_variable_bounds(lp, *args):
    return lp.change_variable_bounds(*args)


def change_variable_objective(lp, *args):
    return lp.change_variable_objective(*args)


def change_coefficient(lp, *args):
    return lp.change_coefficient(*args)


def set_parameter(lp, parameter_name, value):
    return lp.change_parameter(parameter_name, value)


def solve_problem(lp, **kwargs):
    return lp.solve_problem(**kwargs)


def get_status(lp):
    return lp.get_status()


def get_objective_value(lp):
    return lp.get_objective_value()


def format_solution(lp, cobra_model):
    return lp.format_solution(cobra_model)



import re
from uuid import uuid4
from warnings import warn

from numpy import array, object as np_object
from scipy.io import loadmat, savemat
from scipy.sparse import coo_matrix

from .. import Model, Metabolite, Reaction

# try to use an ordered dict
try:
    from scipy.version import short_version
    scipy_version = int(short_version.split(".")[1])
    # if scipy version is earlier than 0.11, OrderedDict will not work, so use
    # a dict instead
    if scipy_version < 11:
        dicttype = dict
    else:
        from collections import OrderedDict as dicttype
    del short_version, scipy_version
except ImportError:
    dicttype = dict


# precompiled regular expressions
_bracket_re = re.compile("r\[[a-z]\]$")
_underscore_re = re.compile(r"_[a-z]$")


def _get_id_comparment(id):
    """extract the compartment from the id string"""
    bracket_search = _bracket_re.findall(id)
    if len(bracket_search) == 1:
        return bracket_search[0][1]
    underscore_search = _underscore_re.findall(id)
    if len(underscore_search) == 1:
        return underscore_search[0][1]
    return None


def _cell(x):
    """translate an array x into a MATLAB cell array"""
    x_no_none = [i if i is not None else "" for i in x]
    return array(x_no_none, dtype=np_object)


def load_matlab_model(infile_path, variable_name=None):
    """Load a cobra model stored as a .mat file

    infile_path : str

    variable_name : str, optional
        The variable name of the model in the .mat file. If this is not
        specified, then the first MATLAB variable which looks like a COBRA
        model will be used

    """
    data = loadmat(infile_path)
    if variable_name is None:
        # skip meta variables
        meta_vars = {"__globals__", "__header__", "__version__"}
        possible_names = sorted(i for i in data if i not in meta_vars)
        if len(possible_names) == 1:
            variable_name = possible_names[0]
    if variable_name is not None:
        return from_mat_struct(data[variable_name], model_id=variable_name)
    for possible_name in possible_names:
        try:
            return from_mat_struct(data[possible_name], model_id=possible_name)
        except ValueError:
            None
    # If code here is executed, then no model was found.
    raise Exception("no COBRA model found")


def save_matlab_model(model, file_name, varname=None):
    """Save the cobra model as a .mat file.

    This .mat file can be used directly in the MATLAB version of COBRA.

    model : :class:`~cobra.core.Model.Model` object

    file_name : str or file-like object

    """
    if varname is None:
        varname = str(model.id) \
            if model.id is not None and len(model.id) > 0 \
            else "exported_model"
    mat = create_mat_dict(model)
    savemat(file_name, {varname: mat},
            appendmat=True, oned_as="column")


def create_mat_dict(model):
    """create a dict mapping model attributes to arrays"""
    model = model.to_array_based_model()
    rxns = model.reactions
    mets = model.metabolites
    mat = dicttype()
    mat["mets"] = _cell(mets.list_attr("id"))
    mat["metNames"] = _cell(mets.list_attr("name"))
    mat["metFormulas"] = _cell([str(m.formula) for m in mets])
    try:
        mat["metCharge"] = array(mets.list_attr("charge")) * 1.
    except TypeError:
        # can't have any None entries for charge, or this will fail
        pass
    mat["genes"] = _cell(model.genes.list_attr("id"))
    mat["grRules"] = _cell(rxns.list_attr("gene_reaction_rule"))
    mat["rxns"] = _cell(rxns.list_attr("id"))
    mat["rxnNames"] = _cell(rxns.list_attr("name"))
    mat["subSystems"] = _cell(rxns.list_attr("subsystem"))
    mat["csense"] = "".join(model._constraint_sense)
    mat["S"] = model.S if model.S is not None else [[]]
    # multiply by 1 to convert to float, working around scipy bug
    # https://github.com/scipy/scipy/issues/4537
    mat["lb"] = array(rxns.list_attr("lower_bound")) * 1.
    mat["ub"] = array(rxns.list_attr("upper_bound")) * 1.
    mat["b"] = array(mets.list_attr("_bound")) * 1.
    mat["c"] = array(rxns.list_attr("objective_coefficient")) * 1.
    mat["rev"] = array(rxns.list_attr("reversibility")) * 1
    mat["description"] = str(model.id)
    return mat


def from_mat_struct(mat_struct, model_id=None):
    """create a model from the COBRA toolbox struct

    The struct will be a dict read in by scipy.io.loadmat

    """
    m = mat_struct
    if m.dtype.names is None:
        raise ValueError("not a valid mat struct")
    if not set(["rxns", "mets", "S", "lb", "ub"]) <= set(m.dtype.names):
        raise ValueError("not a valid mat struct")
    if "c" in m.dtype.names:
        c_vec = m["c"][0, 0]
    else:
        c_vec = None
        warn("objective vector 'c' not found")
    model = Model()
    if "description" in m:
        model.id = m["description"][0, 0][0]
    elif model_id is not None:
        model.id = model_id
    else:
        model.id = "imported_model"
    for i, name in enumerate(m["mets"][0, 0]):
        new_metabolite = Metabolite()
        new_metabolite.id = str(name[0][0])
        new_metabolite.compartment = _get_id_comparment(new_metabolite.id)
        try:
            new_metabolite.name = str(m["metNames"][0, 0][i][0][0])
        except (IndexError, ValueError):
            pass
        try:
            new_metabolite.formula = str(m["metFormulas"][0][0][i][0][0])
        except (IndexError, ValueError):
            pass
        try:
            new_metabolite.charge = float(m["metCharge"][0, 0][i][0])
            int_charge = int(new_metabolite.charge)
            if new_metabolite.charge == int_charge:
                new_metabolite.charge = int_charge
        except (IndexError, ValueError):
            pass
        model.add_metabolites([new_metabolite])
    new_reactions = []
    for i, name in enumerate(m["rxns"][0, 0]):
        new_reaction = Reaction()
        new_reaction.id = str(name[0][0])
        new_reaction.lower_bound = float(m["lb"][0, 0][i][0])
        new_reaction.upper_bound = float(m["ub"][0, 0][i][0])
        if c_vec is not None:
            new_reaction.objective_coefficient = float(c_vec[i][0])
        try:
            new_reaction.gene_reaction_rule = str(m['grRules'][0, 0][i][0][0])
        except (IndexError, ValueError):
            pass
        try:
            new_reaction.name = str(m["rxnNames"][0, 0][i][0][0])
        except (IndexError, ValueError):
            pass
        try:
            new_reaction.subsystem = str(m['subSystems'][0, 0][i][0][0])
        except (IndexError, ValueError):
            pass
        new_reactions.append(new_reaction)
    model.add_reactions(new_reactions)
    coo = coo_matrix(m["S"][0, 0])
    for i, j, v in zip(coo.row, coo.col, coo.data):
        model.reactions[j].add_metabolites({model.metabolites[i]: v})
    return model


def _check(result):
    """ensure success of a pymatbridge operation"""
    if result["success"] is not True:
        raise RuntimeError(result["content"]["stdout"])


def model_to_pymatbridge(model, variable_name="model", matlab=None):
    """send the model to a MATLAB workspace through pymatbridge

    This model can then be manipulated through the COBRA toolbox

    variable_name: str
        The variable name to which the model will be assigned in the
        MATLAB workspace

    matlab: None or pymatbridge.Matlab instance
        The MATLAB workspace to which the variable will be sent. If
        this is None, then this will be sent to the same environment
        used in IPython magics.

    """
    if matlab is None:  # assumed to be running an IPython magic
        from IPython import get_ipython
        matlab = get_ipython().magics_manager.registry["MatlabMagics"].Matlab
    model_info = create_mat_dict(model)
    S = model_info["S"].todok()
    model_info["S"] = 0
    temp_S_name = "cobra_pymatbridge_temp_" + uuid4().hex
    _check(matlab.set_variable(variable_name, model_info))
    _check(matlab.set_variable(temp_S_name, S))
    _check(matlab.run_code("%s.S = %s;" % (variable_name, temp_S_name)))
    # all vectors need to be transposed
    for i in model_info.keys():
        if i == "S":
            continue
        _check(matlab.run_code("{0}.{1} = {0}.{1}';".format(variable_name, i)))
    _check(matlab.run_code("clear %s;" % temp_S_name))



import re
from warnings import warn
from ast import parse as ast_parse, Name, And, Or, BitOr, BitAnd, \
    BoolOp, Expression, NodeTransformer
from keyword import kwlist

from .Species import Species


keywords = list(kwlist)
keywords.remove("and")
keywords.remove("or")
keywords.extend(("True", "False"))
keyword_re = re.compile(r"(?=\b(%s)\b)" % "|".join(keywords))
number_start_re = re.compile(r"(?=\b[0-9])")

replacements = (
    (".", "__COBRA_DOT__"),
    ("'", "__COBRA_SQUOTE__"),
    ('"', "__COBRA_DQUOTE__"),
    (":", "__COBRA_COLON__"),
    ("/", "__COBRA_FSLASH__"),
    ("\\", "__COBRA_BSLASH"),
    ("-", "__COBRA_DASH__"),
    ("=", "__COBRA_EQ__")
)


# functions for gene reaction rules
def ast2str(expr, level=0, names=None):
    """convert compiled ast to gene_reaction_rule str

    expr: str of a gene reaction rule

    level: internal use only

    names: optional dict of {Gene.id: Gene.name}
        Use this to get a rule str which uses names instead. This
        should be done for display purposes only. All gene_reaction_rule
        strings which are computed with should use the id.
    """
    if isinstance(expr, Expression):
        return ast2str(expr.body, 0, names) \
            if hasattr(expr, "body") else ""
    elif isinstance(expr, Name):
        return names.get(expr.id, expr.id) if names else expr.id
    elif isinstance(expr, BoolOp):
        op = expr.op
        if isinstance(op, Or):
            str_exp = " or ".join(ast2str(i, level + 1, names)
                                  for i in expr.values)
        elif isinstance(op, And):
            str_exp = " and ".join(ast2str(i, level + 1, names)
                                   for i in expr.values)
        else:
            raise TypeError("unsupported operation " + op.__class__.__name)
        return "(" + str_exp + ")" if level else str_exp
    elif expr is None:
        return ""
    else:
        raise TypeError("unsupported operation  " + repr(expr))


def eval_gpr(expr, knockouts):
    """evaluate compiled ast of gene_reaction_rule with knockouts"""
    if isinstance(expr, Expression):
        return eval_gpr(expr.body, knockouts)
    elif isinstance(expr, Name):
        return expr.id not in knockouts
    elif isinstance(expr, BoolOp):
        op = expr.op
        if isinstance(op, Or):
            return any(eval_gpr(i, knockouts) for i in expr.values)
        elif isinstance(op, And):
            return all(eval_gpr(i, knockouts) for i in expr.values)
        else:
            raise TypeError("unsupported operation " + op.__class__.__name__)
    elif expr is None:
        return True
    else:
        raise TypeError("unsupported operation  " + repr(expr))


class GPRCleaner(NodeTransformer):
    """Parses compiled ast of a gene_reaction_rule and identifies genes

    Parts of the tree are rewritten to allow periods in gene ID's and
    bitwise boolean operations"""
    def __init__(self):
        NodeTransformer.__init__(self)
        self.gene_set = set()

    def visit_Name(self, node):
        if node.id.startswith("__cobra_escape__"):
            node.id = node.id[16:]
        for char, escaped in replacements:
            if escaped in node.id:
                node.id = node.id.replace(escaped, char)
        self.gene_set.add(node.id)
        return node

    def visit_BinOp(self, node):
        self.generic_visit(node)
        if isinstance(node.op, BitAnd):
            return BoolOp(And(), (node.left, node.right))
        elif isinstance(node.op, BitOr):
            return BoolOp(Or(), (node.left, node.right))
        else:
            raise TypeError("unsupported operation '%s'" %
                            node.op.__class__.__name__)


def parse_gpr(str_expr):
    """parse gpr into AST

    returns: (ast_tree, {gene_ids})"""
    str_expr = str_expr.strip()
    if len(str_expr) == 0:
        return None, set()
    for char, escaped in replacements:
        if char in str_expr:
            str_expr = str_expr.replace(char, escaped)
    escaped_str = keyword_re.sub("__cobra_escape__", str_expr)
    escaped_str = number_start_re.sub("__cobra_escape__", escaped_str)
    tree = ast_parse(escaped_str, "<string>", "eval")
    cleaner = GPRCleaner()
    cleaner.visit(tree)
    eval_gpr(tree, set())  # ensure the rule can be evaluated
    return tree, cleaner.gene_set


class Gene(Species):

    def __init__(self, id=None, name="", functional=True):
        """
        id: A string.

        name: String.  A human readable name.

        functional: Boolean.  Indicate whether the gene is functional.  If it
        is not functional then it cannot be used in an enzyme complex nor
        can its products be used.

        """
        Species.__init__(self, id=id, name=name)
        self.functional = functional

    def remove_from_model(self, model=None,
                          make_dependent_reactions_nonfunctional=True):
        """Removes the association

        make_dependent_reactions_nonfunctional: Boolean.  If True then replace
        the gene with 'False' in the gene association, else replace the gene
        with 'True'

        .. deprecated :: 0.4
            Use cobra.manipulation.delete_model_genes to simulate knockouts
            and cobra.manipulation.remove_genes to remove genes from
            the model.

        """
        warn("Use cobra.manipulation.remove_genes instead")
        if model is not None:
            if model != self._model:
                raise Exception("%s is a member of %s, not %s" %
                                (repr(self), repr(self._model), repr(model)))
        if self._model is None:
            raise Exception('%s is not in a model' % repr(self))

        if make_dependent_reactions_nonfunctional:
            gene_state = 'False'
        else:
            gene_state = 'True'
        the_gene_re = re.compile('(^|(?<=( |\()))%s(?=( |\)|$))' %
                                 re.escape(self.id))

        self._model.genes.remove(self)
        self._model = None

        for the_reaction in list(self._reaction):
            the_reaction._gene_reaction_rule = the_gene_re.sub(
                gene_state, the_reaction.gene_reaction_rule)
            the_reaction._genes.remove(self)
            # Now, deactivate the reaction if its gene association evaluates
            # to False
            the_gene_reaction_relation = the_reaction.gene_reaction_rule
            for other_gene in the_reaction._genes:
                other_gene_re = re.compile('(^|(?<=( |\()))%s(?=( |\)|$))' %
                                           re.escape(other_gene.id))
                the_gene_reaction_relation = other_gene_re.sub(
                    'True',
                    the_gene_reaction_relation)

            if not eval(the_gene_reaction_relation):
                the_reaction.lower_bound = 0
                the_reaction.upper_bound = 0
        self._reaction.clear()



class Solution(object):
    """Stores the solution from optimizing a cobra.Model. This is
    used to provide a single interface to results from different
    solvers that store their values in different ways.

    f: The objective value

    solver: A string indicating which solver package was used.

    x: List or Array of the values from the primal.

    x_dict: A dictionary of reaction ids that maps to the primal values.

    y: List or Array of the values from the dual.

    y_dict: A dictionary of reaction ids that maps to the dual values.

    """

    def __init__(self, f, x=None,
                 x_dict=None, y=None, y_dict=None,
                 solver=None, the_time=0, status='NA'):
        self.solver = solver
        self.f = f
        self.x = x
        self.x_dict = x_dict
        self.status = status
        self.y = y
        self.y_dict = y_dict

    def dress_results(self, model):
        """.. warning :: deprecated"""
        from warnings import warn
        warn("unnecessary to call this deprecated function")

    def __repr__(self):
        if self.f is None:
            return "<Solution '%s' at 0x%x>" % (self.status, id(self))
        return "<Solution %.2f at 0x%x>" % (self.f, id(self))



from six import iteritems


def add_SBO(model):
    """adds SBO terms for demands and exchanges

    This works for models which follow the standard convention for
    constructing and naming these reactions.

    The reaction should only contain the single metabolite being exchanged,
    and the id should be EX_metid or DM_metid
    """
    for r in model.reactions:
        # don't annotate already annotated reactions
        if r.annotation.get("SBO"):
            continue
        # only doing exchanges
        if len(r.metabolites) != 1:
            continue
        met_id = list(r._metabolites)[0].id
        if r.id.startswith("EX_") and r.id == "EX_" + met_id:
            r.annotation["SBO"] = "SBO:0000627"
        elif r.id.startswith("DM_") and r.id == "DM_" + met_id:
            r.annotation["SBO"] = "SBO:0000628"



#!/usr/bin/env python
# This script regenerates pickles of cobra Models.  Should be
# performed after updating core classes to prevent subtle bugs.
try:
    from cPickle import load, dump
except:
    from pickle import load, dump

from json import dump as json_dump
from collections import OrderedDict

import cobra
from cobra.version import get_version
from cobra.io import read_sbml_model, write_sbml_model, save_matlab_model, \
    save_json_model
from cobra.io.sbml3 import write_sbml2

# ecoli
ecoli_model = read_sbml_model("iJO1366.xml")
with open("iJO1366.pickle", "wb") as outfile:
    dump(ecoli_model, outfile, protocol=2)

# salmonella
salmonella = read_sbml_model("salmonella.xml")
with open("salmonella.genes", "rb") as infile:
    gene_names = load(infile)
for gene in salmonella.genes:
    gene.name = gene_names[gene.id]
with open("salmonella.media", "rb") as infile:
    salmonella.media_compositions = load(infile)
with open("salmonella.pickle", "wb") as outfile:
    dump(salmonella, outfile, protocol=2)

# create mini model from textbook
textbook = read_sbml_model("textbook.xml.gz")
mini = cobra.Model("mini_textbook")
mini.compartments = textbook.compartments


for r in textbook.reactions:
    if r.id in ("GLCpts", "PGI", "PFK", "FBA", "TPI", "GAPD", "PGK", "PGM",
                "ENO", "PYK", "EX_glc__D_e", "EX_h_e", "H2Ot", "ATPM",
                "PIt2r"):
        mini.add_reaction(r.copy())
mini.reactions.ATPM.upper_bound = mini.reactions.PGI.upper_bound
mini.change_objective("ATPM")  # No biomass

# add in some information from iJO1366
mini.add_reaction(ecoli_model.reactions.LDH_D.copy())
mini.add_reaction(ecoli_model.reactions.EX_lac__D_e.copy())
r = cobra.Reaction("D_LACt2")
mini.add_reaction(r)
r.gene_reaction_rule = ecoli_model.reactions.D__LACt2pp.gene_reaction_rule
r.reaction = ecoli_model.reactions.D__LACt2pp.reaction.replace("_p", "_e")
mini.reactions.GLCpts.gene_reaction_rule = \
    ecoli_model.reactions.GLCptspp.gene_reaction_rule

# adjust bounds
for i in ["ATPM", "D_LACt2", "EX_lac__D_e", "LDH_D"]:
    mini.reactions.get_by_id(i).upper_bound = mini.reactions.PGI.upper_bound
for i in ["D_LACt2", "LDH_D"]:
    mini.reactions.get_by_id(i).lower_bound = mini.reactions.PGI.lower_bound
# set names and annotation
for g in mini.genes:
    try:
        tg = textbook.genes.get_by_id(g.id)
    except KeyError:
        continue
    g.name = tg.name
    g.annotation = tg.annotation
mini.reactions.sort()
mini.genes.sort()
mini.metabolites.sort()
# output to various formats
with open("mini.pickle", "wb") as outfile:
    dump(mini, outfile, protocol=2)
save_matlab_model(mini, "mini.mat")
save_json_model(mini, "mini.json", pretty=True)
write_sbml_model(mini, "mini_fbc2.xml")
write_sbml_model(mini, "mini_fbc2.xml.bz2")
write_sbml_model(mini, "mini_fbc2.xml.gz")
write_sbml2(mini, "mini_fbc1.xml", use_fbc_package=True)
write_sbml_model(mini, "mini_cobra.xml", use_fbc_package=False)

# fva results
fva_result = cobra.flux_analysis.flux_variability_analysis(textbook)
clean_result = OrderedDict()
for key in sorted(fva_result):
    clean_result[key] = {k: round(v, 5) for k, v in fva_result[key].items()}
with open("textbook_fva.json", "w") as outfile:
    json_dump(clean_result, outfile)



try:
    import numpy
except:
    numpy = None

from .essentiality import assess_medium_component_essentiality
from .variability import flux_variability_analysis, find_blocked_reactions
from .single_deletion import single_gene_deletion, single_reaction_deletion
from .parsimonious import optimize_minimal_flux
from .loopless import construct_loopless_model
from .gapfilling import growMatch

if numpy:
    from .double_deletion import double_reaction_deletion, double_gene_deletion
    from .phenotype_phase_plane import calculate_phenotype_phase_plane
else:
    from warnings import warn
    warn("double deletions and phase planes requires numpy")
    del warn
del numpy



from sys import maxsize
from warnings import warn
from six import iteritems

from numpy import array, ndarray
from scipy.sparse import lil_matrix, dok_matrix

from .Model import Model


class ArrayBasedModel(Model):
    """ArrayBasedModel is a class that adds arrays and vectors to
    a cobra.Model to make it easier to perform linear algebra operations.

    """

    def __init__(self, description=None, deepcopy_model=False,
                 matrix_type='scipy.lil_matrix'):
        """
        description: None | String | cobra.Model

        deepcopy_model: Boolean.  If True and description is
        a cobra.Model then make a deepcopy of the Model before
        creating the ArrayBasedModel.

        matrix_type: 'scipy.lil_matrix' or 'scipy.dok_matrix'
            Specifies which type of backend matrix to use for S.

        """
        if deepcopy_model and isinstance(description, Model):
            description = description.copy()
        Model.__init__(self, description)
        self._S = None
        self.matrix_type = matrix_type
        self.update()

    # no setter for S at the moment
    @property
    def S(self):
        """Stoichiometric matrix of the model

        This will be formatted as either :class:`~scipy.sparse.lil_matrix`
        or :class:`~scipy.sparse.dok_matrix`

        """
        return self._S

    @property
    def lower_bounds(self):
        return self._lower_bounds

    @lower_bounds.setter
    def lower_bounds(self, vector):
        self._update_from_vector("lower_bounds", vector)

    @property
    def upper_bounds(self):
        return self._upper_bounds

    @upper_bounds.setter
    def upper_bounds(self, vector):
        self._update_from_vector("upper_bounds", vector)

    @property
    def objective_coefficients(self):
        return self._objective_coefficients

    @objective_coefficients.setter
    def objective_coefficients(self, vector):
        self._update_from_vector("objective_coefficients", vector)

    @property
    def b(self):
        """bounds for metabolites as :class:`numpy.ndarray`"""
        return self._b

    @b.setter
    def b(self, vector):
        self._update_from_vector("b", vector)

    @property
    def constraint_sense(self):
        return self._constraint_sense

    @constraint_sense.setter
    def constraint_sense(self, vector):
        self._update_from_vector("_constraint_sense", vector)

    def copy(self):
        """Provides a partial 'deepcopy' of the Model.  All of the Metabolite,
        Gene, and Reaction objects are created anew but in a faster fashion
        than deepcopy

        """
        the_copy = Model.copy(self)
        the_copy.update()
        return the_copy

    def add_metabolites(self, metabolite_list,
                        expand_stoichiometric_matrix=True):
        """Will add a list of metabolites to the the object, if they do not
        exist and then expand the stochiometric matrix

        metabolite_list: A list of :class:`~cobra.core.Metabolite` objects

        expand_stoichimetric_matrix: Boolean.  If True and self.S is
        not None then it will add rows to self.S.  self.S must be
        created after adding reactions and metabolites to self before
        it can be expanded.  Trying to expand self.S when self only
        contains metabolites is ludacris.

        """
        Model.add_metabolites(self, metabolite_list)
        if self._S is not None and expand_stoichiometric_matrix:
            s_expansion = len(self.metabolites) - self._S.shape[0]
            if s_expansion > 0:
                self._S.resize((self._S.shape[0] + s_expansion,
                                self._S.shape[1]))
        self._update_metabolite_vectors()

    def _update_from_vector(self, attribute, vector):
        """convert from model.reactions = v to model.reactions[:] = v"""
        # this will fail if vector is the wrong length
        getattr(self, attribute)[:] = vector

    def _update_reaction(self, reaction):
        """Updates everything associated with the reaction.id of reaction.

        reaction: A cobra.Reaction object, or a list of these objects.

        WARNING: This function is only used after the Model has been
        converted to matrices.  It is typically faster to access the objects
        in the Model directly.  This function will eventually moved to another
        module for advanced users due to the potential for mistakes.


        """
        if not hasattr(reaction, '__iter__'):
            reaction = [reaction]
        Model._update_reaction(reaction)
        for the_reaction in reaction:
            try:
                reaction_index = self.reactions.index(the_reaction.id)
            except KeyError:
                warn(the_reaction.id + ' is not in the model')
                continue

            # zero reaction stoichiometry column
            the_column = self._S[:, reaction_index]
            for nonzero_index in the_column.nonzero()[0]:
                the_column[nonzero_index, 0] = 0
            self._lower_bounds[reaction_index] = the_reaction.lower_bound
            self._upper_bounds[reaction_index] = the_reaction.upper_bound
            self.objective_coefficients[
                reaction_index] = the_reaction.objective_coefficient
            self.add_metabolites(the_reaction._metabolites)
            # Make sure that the metabolites are the ones contained in the
            # model
            the_reaction._metabolites = [self.metabolite.get_by_id(x.id)
                                         for x in the_reaction._metabolites]
            # Update the stoichiometric matrix
            metabolite_indices = map(
                self.metabolites.index,
                the_reaction._metabolites)
            for (index, metabolite_index) in enumerate(metabolite_indices):
                self._S[metabolite_index, reaction_index] = \
                    the_reaction.stoichiometric_coefficients[index]

    def add_reactions(self, reaction_list, update_matrices=True):
        """Will add a cobra.Reaction object to the model, if
        reaction.id is not in self.reactions.

        reaction_list: A :class:`~cobra.core.Reaction` object or a list of them

        update_matrices:  Boolean.  If true populate / update matrices
        S, lower_bounds, upper_bounds, .... Note this is slow to run
        for very large models and using this option with repeated calls
        will degrade performance.  Better to call self.update() after
        adding all reactions.


         If the stoichiometric matrix is initially empty then initialize a 1x1
         sparse matrix and add more rows as needed in the self.add_metabolites
         function

        """
        Model.add_reactions(self, reaction_list)
        if update_matrices:
            self._update_matrices(reaction_list)

    def remove_reactions(self, reactions, update_matrices=True, **kwargs):
        """remove reactions from the model

        See :func:`cobra.core.Model.Model.remove_reactions`

        update_matrices:  Boolean
            If true populate / update matrices S, lower_bounds, upper_bounds.
            Note that this is slow to run for very large models, and using this
            option with repeated calls will degrade performance.

        """
        Model.remove_reactions(self, reactions, **kwargs)
        if update_matrices:
            self._update_matrices()

    def _construct_matrices(self):
        """Large sparse matrices take time to construct and to read / write.
        This function allows one to let the model exists without cobra_model.S
        and then generate it at needed.

        """
        self._update_matrices()  # This does basic construction as well.

    def _update_reaction_vectors(self):
        """regenerates the lower_bounds, upper_bounds,
        and objective_coefficients vectors.

        WARNING: This function is only used after the Model has been
        converted to matrices.  It is typically faster to access the objects
        in the Model directly.  This function will eventually moved to another
        module for advanced users due to the potential for mistakes.


        """
        self._lower_bounds = LinkedArray(self.reactions, "lower_bound")
        self._upper_bounds = LinkedArray(self.reactions, "upper_bound")
        self._objective_coefficients = LinkedArray(self.reactions,
                                                   "objective_coefficient")

    def _update_metabolite_vectors(self):
        """regenerates _b and _constraint_sense

        WARNING: This function is only used after the Model has been
        converted to matrices.  It is typically faster to access the objects
        in the Model directly.  This function will eventually moved to another
        module for advanced users due to the potential for mistakes.

        """
        self._b = LinkedArray(self.metabolites, "_bound")
        self._constraint_sense = LinkedArray(
            self.metabolites,
            "_constraint_sense")

    def _update_matrices(self, reaction_list=None):
        """
        reaction_list: None or a list of cobra.Reaction objects that are in
        self.reactions.  If None then reconstruct the whole matrix.

        NOTE: reaction_list is assumed to be at the end of self.reactions.

        In the future, we'll be able to use reactions from anywhere in the
        list

        WARNING: This function is only used after the Model has been
        converted to matrices.  It is typically faster to access the objects
        in the Model directly.  This function will eventually moved to another
        module for advanced users due to the potential for mistakes.

        """
        # no need to create matrix if there are no reactions or metabolites
        if len(self.reactions) == 0 and len(self.metabolites) == 0:
            return
        elif len(self.metabolites) == 0:
            self._update_reaction_vectors()
            return
        elif len(self.reactions) == 0:
            self._update_metabolite_vectors()
            return
        # Pretty much all of these things are unnecessary to use the objects
        # and interact with the optimization solvers.  It might be best to move
        # them to linear algebra modules.  If no reactions are present in the
        # Model, initialize the arrays
        if self._S is None or reaction_list is None:
            reaction_list = self.reactions
            SMatrix = SMatrix_classes[self.matrix_type]
            self._S = SMatrix((len(self.metabolites),
                               len(self.reactions)), model=self)
            self._update_reaction_vectors()
        else:  # Expand the arrays to accomodate the new reaction
            self._S.resize((len(self.metabolites),
                            len(self.reactions)))
            lower_bounds = array([x.lower_bound
                                  for x in reaction_list])
            upper_bounds = array([x.upper_bound
                                  for x in reaction_list])
            objective_coefficients = array([x.objective_coefficient
                                            for x in reaction_list])
            self._lower_bounds._extend(lower_bounds)
            self._upper_bounds._extend(upper_bounds)
            self._objective_coefficients._extend(objective_coefficients)

        coefficient_dictionary = {}
        for the_reaction in reaction_list:
            reaction_index = self.reactions.index(the_reaction.id)
            for the_key, the_value in the_reaction._metabolites.items():
                coefficient_dictionary[(self.metabolites.index(the_key.id),
                                        reaction_index)] = the_value

        self._S.update(coefficient_dictionary)

    def update(self):
        """Regenerates the stoichiometric matrix and vectors"""
        self._update_matrices()
        self._update_metabolite_vectors()


class LinkedArray(ndarray):
    """A :class:`numpy.ndarray` which updates an attribute from a list"""
    def __new__(cls, list, attribute):
        # construct a new ndarray with the values from the list
        # For example, if the list if model.reactions and the attribute is
        # "lower_bound" create an array of [reaction.lower_bound for ... ]
        x = array([getattr(i, attribute) for i in list]).view(cls)
        return x.copy()

    def __init__(self, list, attribute):
        self._list = list
        self._attr = attribute

    def __setitem__(self, index, value):
        ndarray.__setitem__(self, index, value)
        if isinstance(index, slice):
            for i, entry in enumerate(self._list[index]):
                setattr(entry, self._attr, value)
        else:
            setattr(self._list[index], self._attr, value)

    def __setslice__(self, i, j, value):
        ndarray.__setitem__(self, slice(i, j), value)
        if j == maxsize:
            j = len(self)
        if hasattr(value, "__getitem__"):  # setting to a list
            for index in range(i, j):
                setattr(self._list[index], self._attr, value[index])
        else:
            for index in range(i, j):
                setattr(self._list[index], self._attr, value)

    def _extend(self, other):
        old_size = len(self)
        new_size = old_size + len(other)
        self.resize(new_size, refcheck=False)
        ndarray.__setitem__(self, slice(old_size, new_size), other)


class SMatrix_dok(dok_matrix):
    """A 2D sparse dok matrix which maintains links to a cobra Model"""

    def __init__(self, *args, **kwargs):
        dok_matrix.__init__(self, *args)
        self.format = "dok"
        self._model = kwargs["model"] if "model" in kwargs else None

    def __setitem__(self, index, value):
        dok_matrix.__setitem__(self, index, value)
        if isinstance(index[0], int) and isinstance(index[1], int):
            reaction = self._model.reactions[index[1]]
            if value != 0:
                reaction.add_metabolites(
                    {self._model.metabolites[index[0]]: value}, combine=False)
            else:  # setting 0 means metabolites should be removed
                metabolite = self._model.metabolites[index[0]]
                if metabolite in reaction._metabolites:
                    reaction.pop(metabolite)

    def tolil(self):
        new = SMatrix_lil(dok_matrix.tolil(self), model=self._model)
        return new


class SMatrix_lil(lil_matrix):
    """A 2D sparse lil matrix which maintains links to a cobra Model"""

    def __init__(self, *args, **kwargs):
        lil_matrix.__init__(self, *args)
        self.format = "lil"
        self._model = kwargs["model"] if "model" in kwargs else None

    def __setitem__(self, index, value):
        lil_matrix.__setitem__(self, index, value)
        if isinstance(index[0], int):
            metabolites = [self._model.metabolites[index[0]]]
        else:
            metabolites = self._model.metabolites[index[0]]

        if isinstance(index[1], int):
            reactions = [self._model.reactions[index[1]]]
        else:
            reactions = self._model.reactions[index[1]]

        if value == 0:  # remove_metabolites
            met_set = set(metabolites)
            for reaction in reactions:
                to_remove = met_set.intersection(reaction._metabolites)
                for i in to_remove:
                    reaction.pop(i)
        else:  # add metabolites
            met_dict = {met: value for met in metabolites}
            for reaction in reactions:
                reaction.add_metabolites(met_dict, combine=False)

    def update(self, value_dict):
        """update matrix without propagating to model"""
        if len(value_dict) < 100:  # TODO benchmark for heuristic
            for index, value in iteritems(value_dict):
                lil_matrix.__setitem__(self, index, value)
        else:
            matrix = lil_matrix.todok(self)
            matrix.update(value_dict)
            self = SMatrix_lil(matrix.tolil(), model=self._model)
            self._model._S = self

    def todok(self):
        new = SMatrix_dok(lil_matrix.todok(self), model=self._model)
        return new

    # TODO: check if implemented before using own function
    def resize(self, shape):
        matrix = lil_matrix.todok(self)
        matrix.resize(shape)
        self = SMatrix_lil(matrix.tolil(), model=self._model)
        self._model._S = self


SMatrix_classes = {"scipy.dok_matrix": SMatrix_dok,
                   "scipy.lil_matrix": SMatrix_lil}



#inspired by sage/src/sage/numerical/backends/glpk_backend.pxd

cdef extern from "glpk.h":
    ctypedef struct glp_prob "glp_prob":
        pass
    ctypedef struct glp_iocp "glp_iocp":
        int msg_lev
        int br_tech
        int bt_tech
        int pp_tech
        int fp_heur
        int gmi_cuts
        int mir_cuts
        int cov_cuts
        int clq_cuts
        double tol_int
        double tol_obj
        double mip_gap
        int tm_lim
        int out_frq
        int out_dly
        int presolve
        int binarize
    ctypedef struct glp_smcp "glp_smcp":
        int msg_lev
        int meth
        int pricing
        int r_test
        double tol_bnd
        double tol_dj
        double tol_piv
        double obj_ll
        double obj_ul
        int it_lim
        int tm_lim
        int out_frq
        int out_dly
        int presolve
    glp_iocp * new_glp_iocp "new glp_iocp" ()
    void glp_init_iocp(glp_iocp *)
    void glp_init_smcp(glp_smcp *)
    glp_prob * glp_create_prob()
    void glp_set_prob_name(glp_prob *, char *)
    void glp_set_obj_dir(glp_prob *, int)
    void glp_add_rows(glp_prob *, int)
    void glp_add_cols(glp_prob *, int)
    void glp_del_rows(glp_prob *, int, int *)
    void glp_set_row_name(glp_prob *, int, char *)
    void glp_set_col_name(glp_prob *, int, char *)
    void glp_set_row_bnds(glp_prob *, int, int, double, double)
    void glp_set_col_bnds(glp_prob *, int, int, double, double)
    void glp_set_obj_coef(glp_prob *, int, double)
    void glp_load_matrix(glp_prob *, int, int *, int *, double *)
    int glp_simplex(glp_prob *, glp_smcp *)
    int glp_exact(glp_prob *, glp_smcp *)  # requires gmp
    int glp_intopt(glp_prob *, glp_iocp *)
    void glp_std_basis(glp_prob *)
    void glp_delete_prob(glp_prob *)
    double glp_get_col_prim(glp_prob *, int)
    double glp_get_obj_val(glp_prob *)
    double glp_get_col_dual(glp_prob *, int)
    double glp_get_row_dual(glp_prob *, int)
    int glp_print_ranges(glp_prob *lp, int,int, int, char *fname)
    int glp_get_num_rows(glp_prob *)
    int glp_get_num_cols(glp_prob *)
    int glp_get_num_int(glp_prob *)
    double glp_mip_col_val(glp_prob *, int)
    double glp_mip_obj_val(glp_prob *)
    void glp_set_col_kind(glp_prob *, int, int)
    int glp_write_mps(glp_prob *lp, int fmt, void *parm, char *fname)
    int glp_write_lp(glp_prob *lp, void *parm, char *fname)
    int glp_write_prob(glp_prob *P, int flags, char *fname)
    int glp_read_prob(glp_prob *P, int flags, char *fname)

    void glp_set_prob_name(glp_prob *lp, char *name)
    void glp_set_obj_name(glp_prob *lp, char *name)
    void glp_set_row_name(glp_prob *lp, int i, char *name)
    void glp_set_col_name(glp_prob *lp, int i, char *name)

    double glp_get_row_ub(glp_prob *lp, int i)
    double glp_get_row_lb(glp_prob *lp, int i)

    double glp_get_col_ub(glp_prob *lp, int i)
    double glp_get_col_lb(glp_prob *lp, int i)
    void glp_set_col_ub(glp_prob *lp, int i, double value)
    void glp_set_col_lb(glp_prob *lp, int i, double value)


    void glp_create_index(glp_prob *P)
    int glp_find_row(glp_prob *P, const char *name)
    int glp_find_col(glp_prob *P, const char *name)
    void glp_delete_index(glp_prob *P)

    double glp_get_col_lb(glp_prob *lp, int i)
    double glp_get_col_ub(glp_prob *lp, int i)

    void glp_scale_prob(glp_prob *lp, int flags)
    void glp_unscale_prob(glp_prob *lp)

    int glp_get_prim_stat(glp_prob *lp)
    int glp_get_status(glp_prob *lp)
    int glp_mip_status(glp_prob *lp)
    int glp_get_num_nz(glp_prob *lp)
    int glp_set_mat_row(glp_prob *lp, int, int, int *, double * )
    int glp_set_mat_col(glp_prob *lp, int, int, int *, double * )
    int glp_get_mat_row(glp_prob *lp, int, int *, double * )
    int glp_get_mat_col(glp_prob *lp, int, int *, double * )
    double glp_get_row_ub(glp_prob *lp, int)
    double glp_get_row_lb(glp_prob *lp, int)
    int glp_get_col_kind(glp_prob *lp, int)
    double glp_get_obj_coef(glp_prob *lp, int)
    int glp_get_obj_dir(glp_prob *lp)
    void glp_copy_prob(glp_prob *dst, glp_prob *src, int names)

    const char *glp_version()

    # output redirection
    int glp_term_out(int flag)
    void glp_term_hook(int (*func)(void *info, const char *s), void *info)

    int glp_warm_up(glp_prob *P)
    void glp_adv_basis(glp_prob *P, int flags)

    # constants

    # constants for smcp control

    int GLP_MSG_OFF
    int GLP_MSG_ERR
    int GLP_MSG_ON
    int GLP_MSG_ALL

    int GLP_PRIMAL
    int GLP_DUALP
    int GLP_DUAL

    int GLP_PT_STD
    int GLP_PT_PSE

    int GLP_RT_STD
    int GLP_RT_HAR

    double DBL_MAX

    int INT_MAX

    int GLP_ON
    int GLP_OFF

    # constants for scaling the problem
    int GLP_SF_AUTO
    int GLP_SF_GM
    int GLP_SF_EQ
    int GLP_SF_2N
    int GLP_SF_SKIP

    # constants for iocp control, not already in simplex

    int GLP_BR_FFV
    int GLP_BR_LFV
    int GLP_BR_MFV
    int GLP_BR_DTH
    int GLP_BR_PCH

    int GLP_BT_DFS
    int GLP_BT_BFS
    int GLP_BT_BLB
    int GLP_BT_BPH

    int GLP_PP_NONE
    int GLP_PP_ROOT
    int GLP_PP_ALL

    # error codes
    int GLP_EBADB
    int GLP_ESING
    int GLP_ECOND
    int GLP_EBOUND
    int GLP_EFAIL
    int GLP_EOBJLL
    int GLP_EOBJUL
    int GLP_EITLIM
    int GLP_ETMLIM
    int GLP_ENOPFS
    int GLP_ENODFS
    int GLP_EROOT
    int GLP_ESTOP
    int GLP_EMIPGAP
    int GLP_ENOFEAS
    int GLP_ENOCVG
    int GLP_EINSTAB
    int GLP_EDATA
    int GLP_ERANGE


    int GLP_UNDEF
    int GLP_OPT
    int GLP_FEAS
    int GLP_NOFEAS
    int GLP_INFEAS
    int GLP_UNBND

    # other constants

    int GLP_MAX
    int GLP_MIN
    int GLP_UP
    int GLP_FR
    int GLP_DB
    int GLP_FX
    int GLP_LO
    int GLP_CV
    int GLP_IV
    int GLP_BV
    int GLP_MPS_DECK
    int GLP_MPS_FILE

    int GLP_MSG_DBG



from warnings import warn

from six import iteritems, string_types
from ..core.Metabolite import Metabolite
from ..solvers import solver_dict, get_solver_name


def flux_variability_analysis(cobra_model, reaction_list=None,
                              fraction_of_optimum=1.0, solver=None,
                              objective_sense="maximize", **solver_args):
    """Runs flux variability analysis to find max/min flux values

    cobra_model : :class:`~cobra.core.Model`:

    reaction_list : list of :class:`~cobra.core.Reaction`: or their id's
        The id's for which FVA should be run. If this is None, the bounds
        will be comptued for all reactions in the model.

    fraction_of_optimum : fraction of optimum which must be maintained.
        The original objective reaction is constrained to be greater than
        maximal_value * fraction_of_optimum

    solver : string of solver name
        If None is given, the default solver will be used.

    """
    if reaction_list is None and "the_reactions" in solver_args:
        reaction_list = solver_args.pop("the_reactions")
        warn("the_reactions is deprecated. Please use reaction_list=")
    if reaction_list is None:
        reaction_list = cobra_model.reactions
    solver = solver_dict[get_solver_name() if solver is None else solver]
    lp = solver.create_problem(cobra_model)
    solver.solve_problem(lp, objective_sense=objective_sense)
    solution = solver.format_solution(lp, cobra_model)
    if solution.status != "optimal":
        raise ValueError("FVA requires the solution status to be optimal, not "
                         + solution.status)
    # set all objective coefficients to 0
    for i, r in enumerate(cobra_model.reactions):
        if r.objective_coefficient != 0:
            f = solution.x_dict[r.id]
            new_bounds = (f * fraction_of_optimum, f)
            solver.change_variable_bounds(lp, i,
                                          min(new_bounds), max(new_bounds))
            solver.change_variable_objective(lp, i, 0.)
    return calculate_lp_variability(lp, solver, cobra_model, reaction_list,
                                    **solver_args)


def calculate_lp_variability(lp, solver, cobra_model, reaction_list,
                             **solver_args):
    """calculate max and min of selected variables in an LP"""
    fva_results = {}
    for r in reaction_list:
        r_id = str(r)
        i = cobra_model.reactions.index(r_id)
        fva_results[r_id] = {}
        solver.change_variable_objective(lp, i, 1.)
        solver.solve_problem(lp, objective_sense="maximize", **solver_args)
        fva_results[r_id]["maximum"] = solver.get_objective_value(lp)
        solver.solve_problem(lp, objective_sense="minimize", **solver_args)
        fva_results[r_id]["minimum"] = solver.get_objective_value(lp)
        # revert the problem to how it was before
        solver.change_variable_objective(lp, i, 0.)
    return fva_results


def find_blocked_reactions(cobra_model, reaction_list=None,
                           solver=None, zero_cutoff=1e-9,
                           open_exchanges=False, **solver_args):
    """Finds reactions that cannot carry a flux with the current
    exchange reaction settings for cobra_model, using flux variability
    analysis.

    """
    if solver is None:
        solver = get_solver_name()
    if open_exchanges:
        # should not unnecessarily change model
        cobra_model = cobra_model.copy()
        for reaction in cobra_model.reactions:
            if reaction.boundary:
                reaction.lower_bound = min(reaction.lower_bound, -1000)
                reaction.upper_bound = max(reaction.upper_bound, 1000)
    if reaction_list is None:
        reaction_list = cobra_model.reactions
    # limit to reactions which are already 0. If the reactions alread
    # carry flux in this solution, then they can not be blocked.
    solution = solver_dict[solver].solve(cobra_model, **solver_args)
    reaction_list = [i for i in reaction_list
                     if abs(solution.x_dict[i.id]) < zero_cutoff]
    # run fva to find reactions where both max and min are 0
    flux_span_dict = flux_variability_analysis(
        cobra_model, fraction_of_optimum=0., reaction_list=reaction_list,
        solver=solver, **solver_args)
    return [k for k, v in iteritems(flux_span_dict)
            if max(map(abs, v.values())) < zero_cutoff]



from copy import deepcopy
from warnings import warn
from itertools import chain
from ast import NodeTransformer

from six import iteritems

from .. import Reaction, Metabolite
from .delete import get_compiled_gene_reaction_rules
from ..core.Gene import ast2str


_renames = (
    (".", "_DOT_"),
    ("(", "_LPAREN_"),
    (")", "_RPAREN_"),
    ("-", "__"),
    ("[", "_LSQBKT"),
    ("]", "_RSQBKT"),
    (",", "_COMMA_"),
    (":", "_COLON_"),
    (">", "_GT_"),
    ("<", "_LT"),
    ("/", "_FLASH"),
    ("\\", "_BSLASH"),
    ("+", "_PLUS_"),
    ("=", "_EQ_"),
    (" ", "_SPACE_"),
    ("'", "_SQUOT_"),
    ('"', "_DQUOT_"),
)


def _escape_str_id(id_str):
    """make a single string id SBML compliant"""
    for c in ("'", '"'):
        if id_str.startswith(c) and id_str.endswith(c) \
                and id_str.count(c) == 2:
            id_str = id_str.strip(c)
    for char, escaped_char in _renames:
        id_str = id_str.replace(char, escaped_char)
    return id_str


class _GeneRenamer(NodeTransformer):

    def visit_Name(self, node):
        node.id = _escape_str_id(node.id)
        return node


def escape_ID(cobra_model):
    """makes all ids SBML compliant"""
    for x in chain([cobra_model],
                   cobra_model.metabolites,
                   cobra_model.reactions,
                   cobra_model.genes):
        x.id = _escape_str_id(x.id)
    cobra_model.repair()
    gene_renamer = _GeneRenamer()
    for rxn, rule in iteritems(get_compiled_gene_reaction_rules(cobra_model)):
        if rule is not None:
            rxn._gene_reaction_rule = ast2str(gene_renamer.visit(rule))


def initialize_growth_medium(cobra_model, the_medium='MgM',
                             external_boundary_compartment='e',
                             external_boundary_reactions=None,
                             reaction_lower_bound=0.,
                             reaction_upper_bound=1000.,
                             irreversible=False,
                             reactions_to_disable=None):
    """Sets all of the input fluxes to the model to zero and then will
    initialize the input fluxes to the values specified in the_medium if
    it is a dict or will see if the model has a composition dict and use
    that to do the initialization.

    cobra_model: A cobra.Model object.


    the_medium: A string, or a dictionary.
    If a string then the initialize_growth_medium function expects that
    the_model has an attribute dictionary called media_compositions, which is a
    dictionary of dictionaries for various medium compositions.  Where a medium
    composition is a dictionary of external boundary reaction ids for the
    medium components and the external boundary fluxes for each medium
    component.


    external_boundary_compartment: None or a string.
    If not None then it specifies the compartment in which to disable all of
    the external systems boundaries.

    external_boundary_reactions: None or a list of external_boundaries that are
    to have their bounds reset.  This acts in conjunction with
    external_boundary_compartment.


    reaction_lower_bound: Float.  The default value to use for the lower
    bound for the boundary reactions.

    reaction_upper_bound: Float.  The default value to use for the upper
    bound for the boundary.

    irreversible: Boolean.  If the model is irreversible then the medium
    composition is taken as the upper bound

    reactions_to_disable: List of reactions for which the upper and lower
    bounds are disabled.  This is superceded by the contents of
    media_composition

    """
    # Zero all of the inputs to the model
    if hasattr(the_medium, 'keys'):
        medium_composition = the_medium
    else:
        if hasattr(cobra_model, 'media_compositions'):
            if the_medium in cobra_model.media_compositions:
                medium_composition = cobra_model.media_compositions[the_medium]
            else:
                raise Exception("%s is not in the model's media list" %
                                the_medium)
        else:
            raise Exception("the model doesn't have attribute "
                            "media_compositions and the medium is not a dict")
    if external_boundary_reactions is not None:
        if isinstance(external_boundary_reactions[0], str):
            external_boundary_reactions = map(cobra_model.reactions.get_by_id,
                                              external_boundary_reactions)
    elif external_boundary_compartment is None:
            warn("We are initializing the medium without first adjusting all"
                 "external boundary reactions")

    # Select the system_boundary reactions to reset
    if external_boundary_compartment is not None:
        _system_boundaries = dict([(x, x.get_compartments())
                                   for x in cobra_model.reactions
                                   if x.boundary == 'system_boundary'])
        [_system_boundaries.pop(k) for k, v in list(_system_boundaries.items())
         if len(v) == 1 and external_boundary_compartment not in v]
        if external_boundary_reactions is None:
            external_boundary_reactions = _system_boundaries.keys()
        else:
            external_boundary_reactions += _system_boundaries.keys()

    for the_reaction in external_boundary_reactions:
        the_reaction.lower_bound = reaction_lower_bound
        if the_reaction.upper_bound == 0:
            the_reaction.upper_bound = reaction_upper_bound
    # Disable specified reactions
    if reactions_to_disable is not None:
        if isinstance(reactions_to_disable[0], str):
            reactions_to_disable = map(cobra_model.reactions.get_by_id,
                                       reactions_to_disable)
        for the_reaction in reactions_to_disable:
            the_reaction.lower_bound = the_reaction.upper_bound = 0.

    # Update the model inputs based on the_medium
    for the_component in medium_composition.keys():
        the_reaction = cobra_model.reactions.get_by_id(the_component)
        if irreversible:
            the_reaction.upper_bound = medium_composition[the_component]
        else:
            the_reaction.lower_bound = medium_composition[the_component]


def convert_to_irreversible(cobra_model):
    """Split reversible reactions into two irreversible reactions

    These two reactions will proceed in opposite directions. This
    guarentees that all reactions in the model will only allow
    positive flux values, which is useful for some modeling problems.

    cobra_model: A Model object which will be modified in place.

    """
    reactions_to_add = []
    for reaction in cobra_model.reactions:
        # If a reaction is reverse only, the forward reaction (which
        # will be constrained to 0) will be left in the model.
        if reaction.lower_bound < 0:
            reverse_reaction = Reaction(reaction.id + "_reverse")
            reverse_reaction.lower_bound = max(0, -reaction.upper_bound)
            reverse_reaction.upper_bound = -reaction.lower_bound
            reverse_reaction.objective_coefficient = \
                reaction.objective_coefficient * -1
            reaction.lower_bound = max(0, reaction.lower_bound)
            reaction.upper_bound = max(0, reaction.upper_bound)
            # Make the directions aware of each other
            reaction.notes["reflection"] = reverse_reaction.id
            reverse_reaction.notes["reflection"] = reaction.id
            reaction_dict = {k: v * -1
                             for k, v in iteritems(reaction._metabolites)}
            reverse_reaction.add_metabolites(reaction_dict)
            reverse_reaction._model = reaction._model
            reverse_reaction._genes = reaction._genes
            for gene in reaction._genes:
                gene._reaction.add(reverse_reaction)
            reverse_reaction.subsystem = reaction.subsystem
            reverse_reaction._gene_reaction_rule = reaction._gene_reaction_rule
            reactions_to_add.append(reverse_reaction)
    cobra_model.add_reactions(reactions_to_add)


def revert_to_reversible(cobra_model, update_solution=True):
    """This function will convert a reversible model made by
    convert_to_irreversible into a reversible model.

    cobra_model: A cobra.Model which will be modified in place.

    """
    reverse_reactions = [x for x in cobra_model.reactions
                         if "reflection" in x.notes and
                         x.id.endswith('_reverse')]

    # If there are no reverse reactions, then there is nothing to do
    if len(reverse_reactions) == 0:
        return

    update_solution = update_solution and cobra_model.solution is not None \
        and cobra_model.solution.status != "NA"

    if update_solution:
        x_dict = cobra_model.solution.x_dict

    for reverse in reverse_reactions:
        forward_id = reverse.notes.pop("reflection")
        forward = cobra_model.reactions.get_by_id(forward_id)
        forward.lower_bound = -reverse.upper_bound
        if forward.upper_bound == 0:
            forward.upper_bound = -reverse.lower_bound

        # update the solution dict
        if update_solution:
            if reverse.id in x_dict:
                x_dict[forward_id] -= x_dict.pop(reverse.id)

        if "reflection" in forward.notes:
            forward.notes.pop("reflection")

    # Since the metabolites and genes are all still in
    # use we can do this faster removal step.  We can
    # probably speed things up here.
    cobra_model.remove_reactions(reverse_reactions)

    # update the solution vector
    if update_solution:
        cobra_model.solution.x_dict = x_dict
        cobra_model.solution.x = [x_dict[r.id] for r in cobra_model.reactions]


def canonical_form(model, objective_sense='maximize',
                   already_irreversible=False, copy=True):
    """Return a model (problem in canonical_form).

    Converts a minimization problem to a maximization, makes all variables
    positive by making reactions irreversible, and converts all constraints to
    <= constraints.


    model: class:`~cobra.core.Model`. The model/problem to convert.

    objective_sense: str. The objective sense of the starting problem, either
    'maximize' or 'minimize'. A minimization problems will be converted to a
    maximization.

    already_irreversible: bool. If the model is already irreversible, then pass
    True.

    copy: bool. Copy the model before making any modifications.

    """
    if copy:
        model = model.copy()

    if not already_irreversible:
        convert_to_irreversible(model)

    if objective_sense == "minimize":
        # if converting min to max, reverse all the objective coefficients
        for reaction in model.reactions:
            reaction.objective_coefficient = - reaction.objective_coefficient
    elif objective_sense != "maximize":
        raise Exception("Invalid objective sense '%s'. "
                        "Must be 'minimize' or 'maximize'." % objective_sense)

    # convert G and E constraints to L constraints
    for metabolite in model.metabolites:
        if metabolite._constraint_sense == "G":
            metabolite._constraint_sense = "L"
            metabolite._bound = - metabolite._bound
            for reaction in metabolite.reactions:
                coeff = reaction.get_coefficient(metabolite)
                # reverse the coefficient
                reaction.add_metabolites({metabolite: -2 * coeff})
        elif metabolite._constraint_sense == "E":
            # change existing constraint to L
            metabolite._constraint_sense = "L"
            # add new constraint
            new_constr = Metabolite("%s__GE_constraint" % metabolite.id)
            new_constr._constraint_sense = "L"
            new_constr._bound = - metabolite._bound
            for reaction in metabolite.reactions:
                coeff = reaction.get_coefficient(metabolite)
                reaction.add_metabolites({new_constr: -coeff})

    # convert lower bounds to LE constraints
    for reaction in model.reactions:
        if reaction.lower_bound < 0:
            raise Exception("Bounds of irreversible reactions should be >= 0,"
                            " for %s" % reaction.id)
        elif reaction.lower_bound == 0:
            continue
        # new constraint for lower bound
        lb_constr = Metabolite("%s__LB_constraint" % reaction.id)
        lb_constr._constraint_sense = "L"
        lb_constr._bound = - reaction.lower_bound
        reaction.add_metabolites({lb_constr: -1})
        reaction.lower_bound = 0

    return model



from ..core import Model, Reaction, Metabolite
from ..manipulation.modify import convert_to_irreversible, canonical_form

from six import iteritems
from collections import defaultdict
from itertools import chain
from copy import deepcopy


def _add_decision_variable(model, reaction_id):
    """Add an integer decision variable for the given reaction."""
    reaction = model.reactions.get_by_id(reaction_id)
    # add integer variable
    var = Reaction("%s_decision_var" % reaction_id)
    var.lower_bound = 0
    var.upper_bound = 1
    var.variable_kind = "integer"
    var.decision_reaction_id = reaction_id
    model.add_reaction(var)
    # add constraints
    # v <= ub * y  -->  v - ub * y <= 0
    ub_constr = Metabolite("%s_upper_bound" % var.id)
    ub_constr._constraint_sense = "L"
    # v >= lb * y  -->  - v + lb * y <= 0
    lb_constr = Metabolite("%s_lower_bound" % var.id)
    lb_constr._constraint_sense = "L"
    reaction.add_metabolites({lb_constr: - 1,
                              ub_constr:   1})
    var.add_metabolites({lb_constr:   reaction.lower_bound,
                         ub_constr: - reaction.upper_bound})
    return var


def set_up_optknock(model, chemical_objective, knockable_reactions,
                    biomass_objective=None, n_knockouts=5,
                    n_knockouts_required=True, dual_maximum=1000, copy=True):
    """Set up the OptKnock problem described by Burgard et al., 2003:

        Burgard AP, Pharkya P, Maranas CD. Optknock: a bilevel programming
        framework for identifying gene knockout strategies for microbial strain
        optimization.  Biotechnol Bioeng. 2003;84(6):647-57.
        http://dx.doi.org/10.1002/bit.10803.


    model : :class:`~cobra.core.Model` object.

    chemical_objective: str. The ID of the reaction to maximize in the outer
    problem.

    knockable_reactions: [str]. A list of reaction IDs that can be knocked out.

    biomass_objective: str. The ID of the reaction to maximize in the inner
    problem. By default, this is the existing objective function in the passed
    model.

    n_knockouts: int. The number of knockouts allowable.

    n_knockouts_required: bool. Require exactly the number of knockouts
    specified by n_knockouts.

    dual_maximum: float or int. The upper bound for dual variables.

    copy: bool. Copy the model before making any modifications.


    Zachary King 2015

    """

    if copy:
        model = model.copy()

    # add the integer decision variables
    decision_variable_ids = [_add_decision_variable(model, r_id).id
                             for r_id in knockable_reactions]

    # inner problem
    inner_problem = model.copy()
    if biomass_objective:
        found = False
        for reaction in inner_problem.reactions:
            obj = reaction.id == biomass_objective
            reaction.objective_coefficient = 1 if obj else 0
            if obj:
                found = True
        if not found:
            raise Exception("Could not find biomass_objective %s in model" %
                            biomass_objective)

    # dual of inner problem
    inner_dual = dual_problem(inner_problem,
                              integer_vars_to_maintain=decision_variable_ids,
                              already_irreversible=False, copy=False,
                              dual_maximum=dual_maximum)

    # add constraints and variables from inner problem to outer problem
    inner_objectives = {}
    for reaction in inner_dual.reactions:
        inner_objectives[reaction.id] = reaction.objective_coefficient
        reaction.objective_coefficient = 0
        if reaction.id in model.reactions:
            existing_reaction = model.reactions.get_by_id(reaction.id)
            for met, coeff in iteritems(reaction._metabolites):
                if met.id in model.metabolites:
                    existing_reaction.add_metabolites(
                        {model.metabolites.get_by_id(met.id): coeff})
                else:
                    existing_reaction.add_metabolites({deepcopy(met): coeff})
        else:
            model.add_reaction(reaction)

    # constraint to set outer and inner objectives equal, and set chemical
    # objective
    equal_objectives_constr = Metabolite("equal_objectives_constraint")
    equal_objectives_constr._constraint_sense = "E"
    equal_objectives_constr._bound = 0
    for reaction in model.reactions:
        if reaction.objective_coefficient != 0:
            reaction.add_metabolites({equal_objectives_constr:
                                      reaction.objective_coefficient})
        inner_objective = inner_objectives.get(reaction.id, 0)
        if inner_objective:
            reaction.add_metabolites(
                {equal_objectives_constr: - inner_objective})
        # set chemical objective
        reaction.objective_coefficient = 1 \
            if reaction.id == chemical_objective else 0

    # add the n_knockouts constraint
    n_knockouts_constr = Metabolite("n_knockouts_constraint")
    n_knockouts_constr._constraint_sense = "E" if n_knockouts_required else "G"
    n_knockouts_constr._bound = len(decision_variable_ids) - n_knockouts
    for r_id in decision_variable_ids:
        reaction = model.reactions.get_by_id(r_id)
        reaction.add_metabolites({n_knockouts_constr: 1})

    return model


def run_optknock(optknock_problem, solver=None, tolerance_integer=1e-9,
                 **kwargs):
    """Run the OptKnock problem created with set_up_optknock.


    optknock_problem: :class:`~cobra.core.Model` object. The problem generated
    by set_up_optknock.

    solver: str. The name of the preferred solver.

    tolerance_integer: float. The integer tolerance for the MILP.

    **kwargs: Keyword arguments are passed to Model.optimize().


    Zachary King 2015

    """
    solution = optknock_problem.optimize(solver=solver,
                                         tolerance_integer=tolerance_integer,
                                         **kwargs)
    solution.knockouts = []
    for reaction in optknock_problem.reactions:
        if solution.x_dict.get(reaction.id, None) == 0:
            r_id = getattr(reaction, "decision_reaction_id", None)
            if r_id is not None:
                solution.knockouts.append(r_id)
    return solution


# This function will generalize the set_up_optknock code to other MILPs:
# def dual_embed(outer_model, inner_model, ..., objective_sense="maximize",
#                integer_vars_to_maintain=[], already_irreversible=False,
#                copy=True, dual_maximum=1000):
#     """Embed the dual of the inner model within the outer model"""


def dual_problem(model, objective_sense="maximize",
                 integer_vars_to_maintain=[],
                 already_irreversible=False, copy=True, dual_maximum=1000):
    """Return a new model representing the dual of the model.

    Make the problem irreversible, then take the dual. Convert the problem:

        Maximize (c^T)x subject to Ax <= b, x >= 0

    which is something like this in COBRApy:

        Maximize sum(objective_coefficient_j * reaction_j for all j)
            s.t.
            sum(coefficient_i_j * reaction_j for all j) <= metabolite_bound_i
            reaction_j <= upper_bound_j
            reaction_j >= 0

    to the problem:

        Minimize (b^T)w subject to (A^T)w >= c, w >= 0

    which is something like this in COBRApy (S matrix is m x n):

        Minimize sum( metabolite_bound_i * dual_i   for all i ) +
                 sum( upper_bound_j *      dual_m+j for all j ) +
            s.t.
             sum( coefficient_i_j * dual_i for all i ) +
             sum( dual_2m+j' for all j' ) >= objective_coefficient_j
            dual_k >= 0


    Arguments
    ---------

    model : :class:`~cobra.core.Model` object.

    objective_sense: str. The objective sense of the starting problem, either
    'maximize' or 'minimize'. A minimization problems will be converted to a
    maximization before taking the dual. This function always returns a
    minimization problem.

    iteger_vars_to_maintain: [str]. A list of IDs for Boolean integer variables
    to be maintained in the dual problem. See 'Maintaining integer variables'
    below for more details

    already_irreversible: Boolean. If True, then do not convert the model to
    irreversible.

    copy: bool. If True, then make a copy of the model before modifying
    it. This is not necessary if already_irreversible is True.

    dual_maximum: float or int. The upper bound for dual variables.


    Maintaining integer variables
    -----------------------------

    The argument integer_vars_to_maintain can be used to specify certin Boolean
    integer variables that will be maintained in the dual problem. This makes
    it possible to join outer and inner problems in a bi-level MILP. The method
    for maintaining integer variables is described by Tepper and Shlomi, 2010:

        Tepper N, Shlomi T. Predicting metabolic engineering knockout
        strategies for chemical production: accounting for competing pathways.
        Bioinformatics.  2010;26(4):536-43. doi:10.1093/bioinformatics/btp704.

    In COBRApy, this roughly translates to transforming (decision variables p,
    integer constraints o):

        Maximize (c^T)x subject to (A_x)x + (A_y)y <= b, x >= 0

        (1) Maximize sum(objective_coefficient_j * reaction_j for all j)
                s.t.
        (2)     sum(coeff_i_j * reaction_j for all j) +
                sum(decision_coeff_i_j * decision_var_j for all j)
                <= metabolite_bound_i
        (3)     reaction_j <= upper_bound_j
        (4)     reaction_j >= 0

    to the problem:

        Minimize (b - (A_y)y)^T w subject to (A_x^T)w >= c, w >= 0

    which linearizes to (with auxiliary variables z):

        Minimize (b^T)w - { ((A_y)y)^T w with yw --> z }
        subject to (A_x^T)w >= c, linearization constraints, w >= 0
          Linearization constraints: z <= w_max * y, z <= w,
                                     z >= w - w_max * (1 - y), z >= 0

        (5) Minimize sum( metabolite_bound_i *  dual_i            for all i ) +
                      sum( upper_bound_j *      dual_m+j          for all j ) +
                    - sum( decision_coeff_i_j * auxiliary_var_i_j
                          for all combinations i, j )
                s.t.
        (6)   - sum( coefficient_i_j * dual_i for all i ) - dual_m+j
              <= - objective_coefficient_j
        (7)     auxiliary_var_i_j - dual_maximum * decision_var_j          <= 0
        (8)     auxiliary_var_i_j - dual_i                                 <= 0
        (9)   - auxiliary_var_i_j + dual_i + dual_maximum * decision_var_j
              <= dual_maximum
       (10)     dual_maximum >= dual_i            >= 0
       (11)     dual_maximum >= dual_m+j          >= 0
       (12)     dual_maximum >= auxiliary_var_i_j >= 0
       (13)                1 >= decision_var_j    >= 0


    Zachary King 2015

    """

    # convert to canonical form and copy
    model = canonical_form(model, objective_sense=objective_sense,
                           already_irreversible=already_irreversible,
                           copy=copy)

    # new model for the dual
    dual = Model("%s_dual" % model.id)

    # keep track of dual_i
    dual_var_for_met = {}

    # add dual variables for constraints. (2) --> dual_i
    for metabolite in model.metabolites:
        # add constraints based on metabolite constraint sense
        if metabolite._constraint_sense != "L":
            raise Exception("Not a less than or equal constraint: %s"
                            % metabolite.id)

        var = Reaction("%s__dual" % metabolite.id)
        # Without auxiliary variables, the objective coefficient would include
        # integer variables when present. However, we will separate out the
        # integer parts into objective coefficients for auxiliary variables.
        var.objective_coefficient = metabolite._bound  # (5)
        # [dual_vars] >= 0
        var.lower_bound = 0
        var.upper_bound = dual_maximum
        dual.add_reaction(var)
        # remember
        dual_var_for_met[metabolite.id] = var

    # keep track of decision variables (integer_vars_to_maintain) as tuples:
    # (reaction in dual problem, reaction in original problem)
    integer_vars_added = []

    # add constraints and upper bound variables
    for reaction in model.reactions:
        # integer vars to maintain
        if reaction.id in integer_vars_to_maintain:
            # keep these integer variables in the dual, with new transformed
            # constraints
            if (reaction.lower_bound not in [0, 1] or
                    reaction.upper_bound not in [0, 1] or
                    reaction.variable_kind != "integer"):
                raise Exception("Reaction %s from integer_vars_to_maintain is "
                                "not a Boolean integer variable" % reaction.id)
            integer_var = Reaction(reaction.id)
            integer_var.upper_bound = reaction.upper_bound
            integer_var.lower_bound = reaction.lower_bound
            integer_var.variable_kind = reaction.variable_kind
            integer_var.objective_coefficient = 0
            # constraints
            dual.add_reaction(integer_var)
            integer_vars_added.append((integer_var, reaction))

        # other vars
        else:
            # other variables become constraints, (1) to (6)
            constr = Metabolite("%s__dual_constrained_by_c" %
                                reaction.id)  # (6)
            constr._constraint_sense = "L"
            constr._bound = - reaction.objective_coefficient
            for met, coeff in iteritems(reaction._metabolites):
                dual_var = dual_var_for_met[met.id]
                dual_var.add_metabolites({constr: - coeff})

            # upper bound constraints -> variables (3) to (5) and (6)
            var_bound = Reaction("%s__dual_for_upper_bound_constraint" %
                                 reaction.id)  # dual_m+j
            var_bound.objective_coefficient = reaction.upper_bound  # (5)
            # [dual_vars] >= 0
            var_bound.lower_bound = 0
            var_bound.upper_bound = dual_maximum
            # add bound dual variables to dual constraints
            var_bound.add_metabolites({constr: -1})  # (6)
            dual.add_reaction(var_bound)

    # add auxiliary variables
    for integer_var, original_reaction in integer_vars_added:
        for metabolite, coeff in iteritems(original_reaction._metabolites):
            dual_var = dual_var_for_met[metabolite.id]
            # create an auxiliary variable
            aux_var = Reaction("%s__auxiliary__%s" % (integer_var.id,
                                                      dual_var.id))
            aux_var.lower_bound = 0
            aux_var.upper_bound = dual_maximum
            aux_var.objective_coefficient = - coeff
            dual.add_reaction(aux_var)

            # add linearization constraints
            # (7)     auxiliary_var_i_j - dual_maximum * decision_var_j    <= 0
            le_decision_constr = Metabolite("%s__le_decision" % aux_var.id)
            le_decision_constr._constraint_sense = "L"
            le_decision_constr._bound = 0
            aux_var.add_metabolites({le_decision_constr: 1})
            integer_var.add_metabolites({le_decision_constr: - dual_maximum})

            # (8)     auxiliary_var_i_j - dual_i                           <= 0
            le_dual_constr = Metabolite("%s__le_dual" % aux_var.id)
            le_dual_constr._constraint_sense = "L"
            le_dual_constr._bound = 0
            aux_var.add_metabolites({le_dual_constr: 1})
            dual_var.add_metabolites({le_dual_constr: -1})

            # (9)   - auxiliary_var_i_j + dual_i +
            #         dual_maximum * decision_var_j <= dual_maximum
            g_constr = Metabolite("%s__g_dual" % aux_var.id)
            g_constr._constraint_sense = "L"
            g_constr._bound = dual_maximum
            aux_var.add_metabolites({g_constr: -1})
            dual_var.add_metabolites({g_constr: 1})
            integer_var.add_metabolites({g_constr: dual_maximum})

    return dual



##cobra.solvers.glpk_solver
#This script provides wrappers for pyglpk 0.3
from warnings import warn
from copy import deepcopy
try:
    # Import izip for python versions < 3.x
    from itertools import izip as zip
except ImportError:
    pass

from six import iteritems
from glpk import LPX

from ..core.Solution import Solution

solver_name = 'glpk'
_SUPPORTS_MILP = True

# solver specific parameters
variable_kind_dict = {'continuous': float, 'integer': int}
status_dict = {'opt': 'optimal', 'nofeas': 'infeasible', 'unbnd': 'unbounded'}
parameter_defaults = {
    'tolerance_feasibility': 1e-6,
    'tolerance_integer': 1e-9,
    'lp_method': 1
}

METHOD_TYPES = {"auto": 2, "primal": 1, "dual": 3}

def get_status(lp):
    status = lp.status
    if status in status_dict:
        status = status_dict[status]
    else:
        status = 'failed'
    return status

def get_objective_value(lp):
    return lp.obj.value

def format_solution(lp, cobra_model, **kwargs):
    status = get_status(lp)
    if status == 'optimal':
        sol = Solution(lp.obj.value, status=status)
        sol.x = [float(c.primal) for c in lp.cols]
        sol.x_dict = {c.name: c.primal for c in lp.cols}

        # return the duals as well as the primals for LPs
        if lp.kind == float:
            sol.y = [float(c.dual) for c in lp.rows]
            sol.y_dict = {c.name: c.dual for c in lp.rows}
        return sol

    return Solution(None, status=status)

def set_parameter(lp, parameter_name, parameter_value):
    """with pyglpk the parameters are set during the solve phase, with
    the exception of objective sense.

    """
    if parameter_name == 'objective_sense':
        if parameter_value.lower() == 'maximize':
            lp.obj.maximize = True
        elif parameter_value.lower() == 'minimize':
            lp.obj.maximize = False
        else:
            raise ValueError("objective_sense should be "
                             "'maximize' or 'minimize'")
    else:
        # This will be made into an exception in the future
        warn("pyglpk parameters (other than objective_sense) are set "
             "during solve_problem")


def create_problem(cobra_model, **kwargs):
    """Solver-specific method for constructing a solver problem from
    a cobra.Model.  This can be tuned for performance using kwargs

    """
    metabolite_to_index = {r: i for i, r in enumerate(cobra_model.metabolites)}

    lp = LPX()        # Create empty problem instance
    lp.name = 'cobra'     # Assign symbolic name to problem
    lp.rows.add(len(cobra_model.metabolites))
    lp.cols.add(len(cobra_model.reactions))

    for r, the_metabolite in zip(lp.rows, cobra_model.metabolites):
        r.name = the_metabolite.id
        b = float(the_metabolite._bound)
        c = the_metabolite._constraint_sense
        if c == 'E':
            r.bounds = b, b     # Set metabolite to steady state levels
        elif c == 'L':
            r.bounds = None, b
        elif c == 'G':
            r.bounds = b, None
        else:
            raise ValueError("invalid constraint sense")

    objective_coefficients = []
    linear_constraints = []
    for c, the_reaction in zip(lp.cols, cobra_model.reactions):
        c.name = the_reaction.id
        c.kind = variable_kind_dict[the_reaction.variable_kind]
        c.bounds = the_reaction.lower_bound, the_reaction.upper_bound
        objective_coefficients.append(float(the_reaction.objective_coefficient))
        for metabolite, coefficient in iteritems(the_reaction._metabolites):
            metabolite_index = metabolite_to_index[metabolite]
            linear_constraints.append((metabolite_index, c.index, coefficient))

    #Add the new objective coefficients to the problem
    lp.obj[:] = objective_coefficients
    #Need to assign lp.matrix after constructing the whole list
    #linear_constraints.sort()  # if we wanted to be 100% deterministic
    lp.matrix = linear_constraints

    # make sure the objective sense is set in create_problem
    objective_sense = kwargs.get("objective_sense", "maximize")
    set_parameter(lp, "objective_sense", objective_sense)

    return lp


def change_variable_bounds(lp, index, lower_bound, upper_bound):
    lp.cols[index].bounds = (lower_bound, upper_bound)


def change_variable_objective(lp, index, objective):
    lp.obj[index] = objective


def update_problem(lp, cobra_model, **kwargs):
    """A performance tunable method for updating a model problem file

    lp: A gurobi problem object

    cobra_model: the cobra.Model corresponding to 'lp'

    """
    #When reusing the basis only assume that the objective coefficients or bounds can change
    #BUG with changing / unchanging the basis
    index_to_metabolite = dict(zip(range(len(cobra_model.metabolites)),
                                   cobra_model.metabolites))
    index_to_reaction = dict(zip(range(len(cobra_model.reactions)),
                                 cobra_model.reactions))
    reaction_to_index = dict(zip(index_to_reaction.values(),
                                 index_to_reaction.keys()))

    try:
        new_objective = kwargs['new_objective']
    except:
        new_objective = None
    if new_objective is not None:
        objective_coefficients = []
        for c in lp.cols:      # Iterate over all rows
            the_reaction = index_to_reaction[c.index]
            c.name = the_reaction.id
            c.bounds = the_reaction.lower_bound, the_reaction.upper_bound
            objective_coefficients.append(float(the_reaction.objective_coefficient))
            c.kind = variable_kind_dict[the_reaction.variable_kind]
        #Add the new objective coefficients to the problem
        lp.obj[:] = objective_coefficients
    else:
        for c in lp.cols:      # Iterate over all rows
            the_reaction = index_to_reaction[c.index]
            c.name = the_reaction.id
            c.bounds = the_reaction.lower_bound, the_reaction.upper_bound
            c.kind = variable_kind_dict[the_reaction.variable_kind]

def change_coefficient(lp, met_index, rxn_index, value):
    col = lp.cols[rxn_index]
    mat = col.matrix
    for i, entry in enumerate(mat):
        if entry[0] == met_index:
            mat[i] = (met_index, value)
            col.matrix = mat
            return
    # need to append
    mat.append((met_index, value))
    col.matrix = mat


def solve_problem(lp, **kwargs):
    """A performance tunable method for updating a model problem file

    lp: a pyGLPK 0.3 problem

    For pyGLPK it is necessary to provide the following parameters, if they
    are not provided then the default settings will be used: tolerance_feasibility,
    tolerance_integer, lp_method, and objective_sense

    """
    parameters = parameter_defaults.copy()
    parameters.update(kwargs)
    if "quadratic_component" in parameters:
        if parameters.pop('quadratic_component') is not None:
            raise Exception('glpk cannot solve QPs')
    lp_args = {}  # only for lp
    extra_args = {}  # added to both lp and milp
    lp_args["tol_bnd"] = parameters.pop("tolerance_feasibility")
    lp_args["tol_dj"] = lp_args["tol_bnd"]
    method = parameters.pop("lp_method")
    if isinstance(method, int):
        lp_args["meth"] = method
    else:
        lp_args["meth"] = METHOD_TYPES[method]
    if "time_limit" in parameters:
        extra_args["tm_lim"] = int(parameters.pop("time_limit") * 1000)
    if "iteration_limit" in parameters:
        extra_args["it_lim"] = parameters.pop("iteration_limit")
    if "objective_sense" in parameters:
        set_parameter(lp, "objective_sense", parameters.pop("objective_sense"))
    tol_int = parameters.pop("tolerance_integer")
    if len(parameters) > 0:
        raise ValueError("Unknown parameters: " + ", ".join(parameters))
    lp_args.update(extra_args)
    # solve the problem
    lp.simplex(**lp_args)
    if lp.kind == int:
        # For MILPs, it is faster to solve LP then move to MILP
        lp.integer(tol_int=tol_int, **extra_args)
    return get_status(lp)


def solve(cobra_model, **kwargs):
    """Smart interface to optimization solver functions that will convert
    the cobra_model to a solver object, set the parameters, and try multiple
    methods to get an optimal solution before returning the solver object and
    a cobra.Solution (which is attached to cobra_model.solution)

    cobra_model: a cobra.Model

    returns a dict: {'the_problem': solver specific object, 'the_solution':
    cobra.Solution for the optimization problem'}


    """
    #Start out with default parameters and then modify if
    #new onese are provided
    the_parameters = deepcopy(parameter_defaults)
    if kwargs:
        the_parameters.update(kwargs)

    for i in ["new_objective", "update_problem", "the_problem"]:
        if i in the_parameters:
            raise Exception("Option %s removed" % i)
    if 'error_reporting' in the_parameters:
        warn("error_reporting deprecated")

    #Create a new problem
    lp = create_problem(cobra_model, **the_parameters)

    ###Try to solve the problem using other methods if the first method doesn't work
    lp_method = the_parameters['lp_method']
    the_methods = [1, 2, 3]
    if lp_method in the_methods:
        the_methods.remove(lp_method)
    #Start with the user specified method
    the_methods.insert(0, lp_method)
    for the_method in the_methods:
        the_parameters['lp_method'] = the_method
        try:
            status = solve_problem(lp, **the_parameters)
        except Exception as e:
            status = 'failed'
        if status == 'optimal':
            break

    return format_solution(lp, cobra_model)



from cylp.cy import CyClpSimplex
from cylp.py.modeling.CyLPModel import CyLPArray
from cylp.cy.CyCoinPackedMatrix import CyCoinPackedMatrix

solver_name = "coin"
_status_translation = {"primal infeasible": "infeasible",
                       "solution": "optimal"}

_SUPPORTS_MILP = True


class Coin(CyClpSimplex):
    cbc = None

    @property
    def status_(self):
        return self.cbc.status if self.cbc else self.getStatusString()

    @property
    def primalVariableSolution_(self):
        return self.cbc.primalVariableSolution if self.cbc \
            else self.primalVariableSolution

    @property
    def objectiveValue_(self):
        return self.cbc.objectiveValue if self.cbc else self.objectiveValue


def create_problem(cobra_model, objective_sense="maximize", **kwargs):
    m = cobra_model.to_array_based_model()
    lp = Coin()
    v = lp.addVariable("v", len(m.reactions))
    for i, rxn in enumerate(m.reactions):
        if rxn.variable_kind == "integer":
            lp.setInteger(v[i])
    S = m.S
    v.lower = CyLPArray(m.lower_bounds)
    v.upper = CyLPArray(m.upper_bounds)
    inf = float("inf")
    cons = zip(m.b, m.constraint_sense)
    b_l = CyLPArray([-inf if s == "L" else b for b, s in cons])
    b_u = CyLPArray([inf if s == "G" else b for b, s in cons])
    lp.addConstraint(b_u >= S * v >= b_l, "b")
    lp.objectiveCoefficients = CyLPArray(m.objective_coefficients)
    set_parameter(lp, "objective_sense", objective_sense)
    set_parameter(lp, "tolerance_feasibility", 1e-9)
    lp.logLevel = 0
    for key, value in kwargs.items():
        set_parameter(lp, key, value)
    return lp


def solve(cobra_model, **kwargs):
    lp = create_problem(cobra_model)
    for key, value in kwargs.items():
        set_parameter(lp, key, value)
    solve_problem(lp)
    return format_solution(lp, cobra_model)


def set_parameter(lp, parameter_name, value):
    if parameter_name == "objective_sense":
        v = str(value).lower()
        if v == "maximize":
            lp.optimizationDirection = "max"
        elif v == "minimize":
            lp.optimizationDirection = "min"
        else:
            raise ValueError("unknown objective sense '%s'" % value)
    elif parameter_name == "tolerance_feasibility":
        lp.primalTolerance = value
    elif parameter_name == "verbose":
        lp.logLevel = value
    elif parameter_name == "quadratic_component":
        set_quadratic_objective(lp, value)
    else:
        setattr(lp, parameter_name, value)


def solve_problem(lp, **kwargs):
    for key, value in kwargs.items():
        set_parameter(lp, key, value)
    if max(lp.integerInformation):
        lp.cbc = lp.getCbcModel()
        lp.cbc.logLevel = lp.logLevel
        return lp.cbc.branchAndBound()
    else:
        lp.cbc = None
        return lp.primal()


def format_solution(lp, cobra_model):
    Solution = cobra_model.solution.__class__
    status = get_status(lp)
    if status != "optimal":  # todo handle other possible
        return Solution(None, status=status)
    solution = Solution(lp.objectiveValue_, status=status)
    x = lp.primalVariableSolution_["v"].tolist()
    solution.x_dict = {r.id: x[i] for i, r in enumerate(cobra_model.reactions)}
    solution.x = x
    # TODO handle y

    return solution


def get_status(lp):
    status = lp.status_
    return _status_translation.get(status, status)


def get_objective_value(lp):
    return lp.objectiveValue_


def change_variable_bounds(lp, index, lower_bound, upper_bound):
    lp.variablesLower[index] = lower_bound
    lp.variablesUpper[index] = upper_bound


def change_coefficient(lp, met_index, rxn_index, value):
    S = lp.coefMatrix
    S[met_index, rxn_index] = value
    lp.coefMatrix = S


def change_variable_objective(lp, index, value):
    lp.setObjectiveCoefficient(index, value)


def _set_quadratic_objective(lp, quadratic_objective):
    """The quadratic routines in CLP do not yet work for GEMs"""
    if not hasattr(quadratic_objective, "tocoo"):
        raise Exception('quadratic component must have method tocoo')
    coo = quadratic_objective.tocoo()
    matrix = CyCoinPackedMatrix(True, coo.row, coo.col, coo.data)
    lp.loadQuadraticObjective(matrix)



from warnings import warn
from itertools import chain, product

from six import iteritems, string_types
import numpy

from ..solvers import get_solver_name, solver_dict
from ..manipulation.delete import find_gene_knockout_reactions, \
    get_compiled_gene_reaction_rules
from .deletion_worker import CobraDeletionPool, CobraDeletionMockPool

try:
    import scipy
except ImportError:
    moma = None
else:
    from . import moma

try:
    from pandas import DataFrame
except:
    DataFrame = None


# Utility functions
def generate_matrix_indexes(ids1, ids2):
    """map an identifier to an entry in the square result matrix"""
    return {id: index for index, id in enumerate(set(chain(ids1, ids2)))}


def yield_upper_tria_indexes(ids1, ids2, id_to_index):
    """gives the necessary indexes in the upper triangle

    ids1 and ids2 are lists of the identifiers i.e. gene id's or reaction
    indexes to be knocked out. id_to_index maps each identifier to its index
    in the result matrix.

    Note that this does not return indexes for the diagonal. Those have
    to be computed separately."""
    # sets to check for inclusion in o(1)
    id_set1 = set(ids1)
    id_set2 = set(ids2)
    for id1, id2 in product(ids1, ids2):
        # indexes in the result matrix
        index1 = id_to_index[id1]
        index2 = id_to_index[id2]
        # upper triangle
        if index2 > index1:
            yield ((index1, index2), (id1, id2))
        # lower triangle but would be skipped, so return in upper triangle
        elif id2 not in id_set1 or id1 not in id_set2:
            yield((index2, index1), (id2, id1))  # note that order flipped


def _format_upper_triangular_matrix(row_indexes, column_indexes, matrix):
    """reformat the square upper-triangular result matrix

    For example, results may look like this
    [[ A  B  C  D]
     [ -  -  -  -]
     [ -  -  E  F]
     [ -  -  -  G]]
    In this case, the second row was skipped. This means we have
    row_indexes [0, 2, 3] and column_indexes [0, 1, 2, 3]

    First, it will reflect the upper triangle into the lower triangle
    [[ A  B  C  D]
     [ B  -  -  -]
     [ C  -  E  F]
     [ D  -  F  G]]

    Finally, it will remove the missing rows and return
    [[ A  B  C  D]
     [ C  -  E  F]
     [ D  -  F  G]]
    """
    results = matrix.copy()
    # Thse select the indexes for the upper triangle. However, switching
    # the order selects the lower triangle.
    triu1, triu2 = numpy.triu_indices(matrix.shape[0])
    # This makes reflection pretty easy
    results[triu2, triu1] = results[triu1, triu2]
    # Remove the missing rows and  return.
    return results[row_indexes, :][:, column_indexes]


def format_results_frame(row_ids, column_ids, matrix, return_frame=False):
    """format results as a pandas.DataFrame if desired/possible

    Otherwise returns a dict of
    {"x": row_ids, "y": column_ids", "data": result_matrx}"""
    if return_frame and DataFrame:
        return DataFrame(data=matrix, index=row_ids, columns=column_ids)
    elif return_frame and not DataFrame:
        warn("could not import pandas.DataFrame")
    return {"x": row_ids, "y": column_ids, "data": matrix}


def double_deletion(cobra_model, element_list_1=None, element_list_2=None,
                    element_type='gene', **kwargs):
    """Wrapper for double_gene_deletion and double_reaction_deletion

    .. deprecated :: 0.4
        Use double_reaction_deletion and double_gene_deletion
    """
    warn("deprecated - use single_reaction_deletion and single_gene_deletion")
    if element_type == "reaction":
        return double_reaction_deletion(cobra_model, element_list_1,
                                        element_list_2, **kwargs)
    elif element_type == "gene":
        return double_gene_deletion(cobra_model, element_list_1,
                                    element_list_2, **kwargs)
    else:
        raise Exception("unknown element type")


def double_reaction_deletion(cobra_model,
                             reaction_list1=None, reaction_list2=None,
                             method="fba", return_frame=False,
                             solver=None, zero_cutoff=1e-12,
                             **kwargs):
    """sequentially knocks out pairs of reactions in a model

    cobra_model : :class:`~cobra.core.Model.Model`
        cobra model in which to perform deletions

    reaction_list1 : [:class:`~cobra.core.Reaction.Reaction`:] (or their id's)
        Reactions to be deleted. These will be the rows in the result.
        If not provided, all reactions will be used.

    reaction_list2 : [:class:`~cobra.core.Reaction`:] (or their id's)
        Reactions to be deleted. These will be the rows in the result.
        If not provided, reaction_list1 will be used.

    method: "fba" or "moma"
        Procedure used to predict the growth rate

    solver: str for solver name
        This must be a QP-capable solver for MOMA. If left unspecified,
        a suitable solver will be automatically chosen.

    zero_cutoff: float
        When checking to see if a value is 0, this threshold is used.

    return_frame: bool
        If true, formats the results as a pandas.Dataframe. Otherwise
        returns a dict of the form:
        {"x": row_labels, "y": column_labels", "data": 2D matrix}
    """
    # handle arguments which need to be passed on
    if solver is None:
        solver = get_solver_name(qp=(method == "moma"))
        kwargs["solver"] = solver
    kwargs["zero_cutoff"] = zero_cutoff

    # generate other arguments

    # identifiers for reactions are their indexes
    if reaction_list1 is None:
        reaction_indexes1 = range(len(cobra_model.reactions))
    else:
        reaction_indexes1 = [cobra_model.reactions.index(r)
                             for r in reaction_list1]
    if reaction_list2 is None:
        reaction_indexes2 = reaction_indexes1
    else:
        reaction_indexes2 = [cobra_model.reactions.index(r)
                             for r in reaction_list2]
    reaction_to_result = generate_matrix_indexes(reaction_indexes1,
                                                 reaction_indexes2)

    # Determine 0 flux reactions. If an optimal solution passes no flux
    # through the deleted reactions, then we know removing them will
    # not change the solution.
    wt_solution = solver_dict[solver].solve(cobra_model)
    if wt_solution.status == "optimal":
        kwargs["wt_growth_rate"] = wt_solution.f
        kwargs["no_flux_reaction_indexes"] = \
            {i for i, v in enumerate(wt_solution.x) if abs(v) < zero_cutoff}
    else:
        warn("wild-type solution status is '%s'" % wt_solution.status)

    # call the computing functions
    if method == "fba":
        results = _double_reaction_deletion_fba(
            cobra_model, reaction_indexes1, reaction_indexes2,
            reaction_to_result, **kwargs)
    elif method == "moma":
        results = _double_reaction_deletion_moma(
            cobra_model, reaction_indexes1, reaction_indexes2,
            reaction_to_result, **kwargs)
    else:
        raise ValueError("Unknown deletion method '%s'" % method)

    # convert upper triangular matrix to full matrix
    full_result = _format_upper_triangular_matrix(
        [reaction_to_result[i] for i in reaction_indexes1],  # row indexes
        [reaction_to_result[i] for i in reaction_indexes2],  # col indexes
        results)

    # format appropriately with labels
    row_ids = [cobra_model.reactions[i].id for i in reaction_indexes1]
    column_ids = [cobra_model.reactions[i].id for i in reaction_indexes2]
    return format_results_frame(row_ids, column_ids,
                                full_result, return_frame)


def double_gene_deletion(cobra_model,
                         gene_list1=None, gene_list2=None,
                         method="fba", return_frame=False,
                         solver=None, zero_cutoff=1e-12,
                         **kwargs):
    """sequentially knocks out pairs of genes in a model

    cobra_model : :class:`~cobra.core.Model.Model`
        cobra model in which to perform deletions

    gene_list1 : [:class:`~cobra.core.Gene.Gene`:] (or their id's)
        Genes to be deleted. These will be the rows in the result.
        If not provided, all reactions will be used.

    gene_list1 : [:class:`~cobra.core.Gene.Gene`:] (or their id's)
        Genes to be deleted. These will be the rows in the result.
        If not provided, reaction_list1 will be used.

    method: "fba" or "moma"
        Procedure used to predict the growth rate

    solver: str for solver name
        This must be a QP-capable solver for MOMA. If left unspecified,
        a suitable solver will be automatically chosen.

    zero_cutoff: float
        When checking to see if a value is 0, this threshold is used.

    number_of_processes: int for number of processes to use.
        If unspecified, the number of parallel processes to use will be
        automatically determined. Setting this to 1 explicitly disables used
        of the multiprocessing library.

    .. note:: multiprocessing is not supported with method=moma

    return_frame: bool
        If true, formats the results as a pandas.Dataframe. Otherwise
        returns a dict of the form:
        {"x": row_labels, "y": column_labels", "data": 2D matrix}
    """
    # handle arguments which need to be passed on
    if solver is None:
        solver = get_solver_name(qp=(method == "moma"))
        kwargs["solver"] = solver
    kwargs["zero_cutoff"] = zero_cutoff

    # generate other arguments

    # identifiers for genes
    if gene_list1 is None:
        gene_ids1 = cobra_model.genes.list_attr("id")
    else:
        gene_ids1 = [str(i) for i in gene_list1]
    if gene_list2 is None:
        gene_ids2 = gene_ids1
    else:
        gene_ids2 = [str(i) for i in gene_list2]

    # The gene_id_to_result dict will map each gene id to the index
    # in the result matrix.
    gene_id_to_result = generate_matrix_indexes(gene_ids1, gene_ids2)

    # Determine 0 flux reactions. If an optimal solution passes no flux
    # through the deleted reactions, then we know removing them will
    # not change the solution.
    wt_solution = solver_dict[solver].solve(cobra_model)
    if wt_solution.status == "optimal":
        kwargs["wt_growth_rate"] = wt_solution.f
        kwargs["no_flux_reaction_indexes"] = \
            {i for i, v in enumerate(wt_solution.x) if abs(v) < zero_cutoff}
    else:
        warn("wild-type solution status is '%s'" % wt_solution.status)

    if method == "fba":
        result = _double_gene_deletion_fba(cobra_model, gene_ids1, gene_ids2,
                                           gene_id_to_result, **kwargs)
    elif method == "moma":
        result = _double_gene_deletion_moma(cobra_model, gene_ids1, gene_ids2,
                                            gene_id_to_result, **kwargs)
    else:
        raise ValueError("Unknown deletion method '%s'" % method)

    # convert upper triangular matrix to full matrix
    full_result = _format_upper_triangular_matrix(
        [gene_id_to_result[id] for id in gene_ids1],  # row indexes
        [gene_id_to_result[id] for id in gene_ids2],  # col indexes,
        result)

    # format as a Dataframe if required
    return format_results_frame(gene_ids1, gene_ids2,
                                full_result, return_frame)


def _double_reaction_deletion_fba(cobra_model, reaction_indexes1,
                                  reaction_indexes2, reaction_to_result,
                                  solver, number_of_processes=None,
                                  zero_cutoff=1e-15, wt_growth_rate=None,
                                  no_flux_reaction_indexes=set(), **kwargs):
    """compute double reaction deletions using fba

    cobra_model: model

    reaction_indexes1, reaction_indexes2: reaction indexes (used as unique
        identifiers)

    reaction_to_result: maps each reaction identifier to the entry in
        the result matrix

    no_flux_reaction_indexes: set of indexes for reactions in the model
        which carry no flux in an optimal solution. For deletions only in
        this set, the result will beset to wt_growth_rate.

    returns an upper triangular square matrix
    """
    if solver is None:
        solver = get_solver_name()

    # generate the square result matrix
    n_results = len(reaction_to_result)
    results = numpy.empty((n_results, n_results))
    results.fill(numpy.nan)

    PoolClass = CobraDeletionMockPool if number_of_processes == 1 \
        else CobraDeletionPool  # explicitly disable multiprocessing

    with PoolClass(cobra_model, n_processes=number_of_processes,
                   solver=solver, **kwargs) as pool:

        # precompute all single deletions in the pool and store them along
        # the diagonal
        for reaction_index, result_index in iteritems(reaction_to_result):
            pool.submit((reaction_index, ), label=result_index)
        for result_index, value in pool.receive_all():
            # if singly lethal, set everything in row and column to 0
            value = value if abs(value) > zero_cutoff else 0.
            if value == 0.:
                results[result_index, :] = 0.
                results[:, result_index] = 0.
            else:  # only the diagonal needs to be set
                results[result_index, result_index] = value

        # Run double knockouts in the upper triangle
        index_selector = yield_upper_tria_indexes(
            reaction_indexes1, reaction_indexes2, reaction_to_result)
        for result_index, (r1_index, r2_index) in index_selector:
            # skip if the result was already computed to be lethal
            if results[result_index] == 0:
                continue
            # reactions removed carry no flux
            if r1_index in no_flux_reaction_indexes and \
                    r2_index in no_flux_reaction_indexes:
                results[result_index] = wt_growth_rate
                continue
            pool.submit((r1_index, r2_index), label=result_index)
        # get results
        for result in pool.receive_all():
            results[result[0]] = result[1]

    return results


def _double_gene_deletion_fba(cobra_model, gene_ids1, gene_ids2,
                              gene_id_to_result, solver,
                              number_of_processes=None, zero_cutoff=1e-12,
                              wt_growth_rate=None,
                              no_flux_reaction_indexes=set(), **kwargs):
    """compute double gene deletions using fba

    cobra_model: model

    gene_ids1, gene_ids2: lists of id's to be knocked out

    gene_id_to_result: maps each gene identifier to the entry in
        the result matrix

    no_flux_reaction_indexes: set of indexes for reactions in the model
        which carry no flux in an optimal solution. For deletions only in
        this set, the result will beset to wt_growth_rate.

    returns an upper triangular square matrix
    """
    # Because each gene reaction rule will be evaluated multiple times
    # the reaction has multiple associated genes being deleted, compiling
    # the gene reaction rules ahead of time increases efficiency greatly.
    compiled_rules = get_compiled_gene_reaction_rules(cobra_model)

    n_results = len(gene_id_to_result)
    results = numpy.empty((n_results, n_results))
    results.fill(numpy.nan)

    if number_of_processes == 1:  # explicitly disable multiprocessing
        PoolClass = CobraDeletionMockPool
    else:
        PoolClass = CobraDeletionPool
    with PoolClass(cobra_model, n_processes=number_of_processes,
                   solver=solver, **kwargs) as pool:
        # precompute all single deletions in the pool and store them along
        # the diagonal
        for gene_id, gene_result_index in iteritems(gene_id_to_result):
            ko_reactions = find_gene_knockout_reactions(
                cobra_model, (cobra_model.genes.get_by_id(gene_id),))
            ko_indexes = [cobra_model.reactions.index(i) for i in ko_reactions]
            pool.submit(ko_indexes, label=gene_result_index)
        for result_index, value in pool.receive_all():
            # if singly lethal, set everything in row and column to 0
            value = value if abs(value) > zero_cutoff else 0.
            if value == 0.:
                results[result_index, :] = 0.
                results[:, result_index] = 0.
            else:  # only the diagonal needs to be set
                results[result_index, result_index] = value

        # Run double knockouts in the upper triangle
        index_selector = yield_upper_tria_indexes(gene_ids1, gene_ids2,
                                                  gene_id_to_result)
        for result_index, (gene1, gene2) in index_selector:

            # if singly lethal the results have already been set
            if results[result_index] == 0:
                continue
            ko_reactions = find_gene_knockout_reactions(
                cobra_model, (gene1, gene2), compiled_rules)
            ko_indexes = [cobra_model.reactions.index(i)
                          for i in ko_reactions]
            # if all removed gene indexes carry no flux
            if len(set(ko_indexes) - no_flux_reaction_indexes) == 0:
                results[result_index] = wt_growth_rate
                continue
            pool.submit(ko_indexes, label=result_index)

        for result in pool.receive_all():
            value = result[1]
            if value < zero_cutoff:
                value = 0
            results[result[0]] = value

    return results


def _double_reaction_deletion_moma(cobra_model, reaction_indexes1,
                                   reaction_indexes2, reaction_to_result,
                                   solver, number_of_processes=1,
                                   zero_cutoff=1e-15, wt_growth_rate=None,
                                   no_flux_reaction_indexes=set(), **kwargs):
    """compute double reaction deletions using moma

    cobra_model: model

    reaction_indexes1, reaction_indexes2: reaction indexes (used as unique
        identifiers)

    reaction_to_result: maps each reaction identifier to the entry in
        the result matrix

    no_flux_reaction_indexes: set of indexes for reactions in the model
        which carry no flux in an optimal solution. For deletions only in
        this set, the result will beset to wt_growth_rate.

    number_of_processes: must be 1. Parallel MOMA not yet implmemented

    returns an upper triangular square matrix
    """
    if number_of_processes > 1:
        raise NotImplementedError("parallel MOMA not implemented")
    if moma is None:
        raise RuntimeError("scipy required for MOMA")

    # generate the square result matrix
    n_results = len(reaction_to_result)
    results = numpy.empty((n_results, n_results))
    results.fill(numpy.nan)

    # function to compute reaction knockouts with moma
    moma_model, moma_obj = moma.create_euclidian_moma_model(cobra_model)

    def run(indexes):
        # If all the reactions carry no flux, deletion will have no effect.
        if no_flux_reaction_indexes.issuperset(indexes):
            return wt_growth_rate
        return moma.moma_knockout(moma_model, moma_obj, indexes,
                                  solver=solver, **kwargs).f

    # precompute all single deletions and store them along the diagonal
    for reaction_index, result_index in iteritems(reaction_to_result):
        value = run((reaction_index,))
        value = value if abs(value) > zero_cutoff else 0.
        results[result_index, result_index] = value
        # if singly lethal, the entire row and column are set to 0
        if value == 0.:
            results[result_index, :] = 0.
            results[:, result_index] = 0.

    # Run double knockouts in the upper triangle
    index_selector = yield_upper_tria_indexes(
        reaction_indexes1, reaction_indexes2, reaction_to_result)
    for result_index, (r1_index, r2_index) in index_selector:
        # skip if the result was already computed to be lethal
        if results[result_index] == 0:
            continue
        else:
            results[result_index] = run((r1_index, r2_index))

    return results


def _double_gene_deletion_moma(cobra_model, gene_ids1, gene_ids2,
                               gene_id_to_result, solver,
                               number_of_processes=1,
                               zero_cutoff=1e-12, wt_growth_rate=None,
                               no_flux_reaction_indexes=set(), **kwargs):
    """compute double gene deletions using moma

    cobra_model: model

    gene_ids1, gene_ids2: lists of id's to be knocked out

    gene_id_to_result: maps each gene identifier to the entry in
        the result matrix

    number_of_processes: must be 1. Parallel MOMA not yet implemented

    no_flux_reaction_indexes: set of indexes for reactions in the model
        which carry no flux in an optimal solution. For deletions only in
        this set, the result will beset to wt_growth_rate.

    returns an upper triangular square matrix
    """
    if number_of_processes > 1:
        raise NotImplementedError("parallel MOMA not implemented")
    if moma is None:
        raise RuntimeError("scipy required for MOMA")

    # Because each gene reaction rule will be evaluated multiple times
    # the reaction has multiple associated genes being deleted, compiling
    # the gene reaction rules ahead of time increases efficiency greatly.
    compiled_rules = get_compiled_gene_reaction_rules(cobra_model)

    # function to compute reaction knockouts with moma
    moma_model, moma_obj = moma.create_euclidian_moma_model(cobra_model)

    def run(gene_ids):
        ko_reactions = find_gene_knockout_reactions(cobra_model, gene_ids)
        ko_indexes = map(cobra_model.reactions.index, ko_reactions)
        # If all the reactions carry no flux, deletion will have no effect.
        if no_flux_reaction_indexes.issuperset(gene_ids):
            return wt_growth_rate
        return moma.moma_knockout(moma_model, moma_obj, ko_indexes,
                                  solver=solver, **kwargs).f

    n_results = len(gene_id_to_result)
    results = numpy.empty((n_results, n_results))
    results.fill(numpy.nan)

    # precompute all single deletions and store them along the diagonal
    for gene_id, result_index in iteritems(gene_id_to_result):
        value = run((gene_id,))
        value = value if abs(value) > zero_cutoff else 0.
        results[result_index, result_index] = value
        # If singly lethal, the entire row and column are set to 0.
        if value == 0.:
            results[result_index, :] = 0.
            results[:, result_index] = 0.

    # Run double knockouts in the upper triangle
    index_selector = yield_upper_tria_indexes(gene_ids1, gene_ids2,
                                              gene_id_to_result)
    for result_index, (gene1, gene2) in index_selector:
        # if singly lethal the results have already been set
        if results[result_index] == 0:
            continue
        results[result_index] = run((gene1, gene2))

    return results



import re
from itertools import islice
from six import string_types, iteritems, PY3

try:
    from numpy import bool_
except:
    bool_ = bool


class DictList(list):
    """A combined dict and list

    This object behaves like a list, but has the O(1) speed
    benefits of a dict when looking up elements by their id.

    """

    def __init__(self, *args):
        if len(args) > 2:
            raise TypeError("takes at most 1 argument (%d given)" % len(args))
        list.__init__(self)
        self._dict = {}
        if len(args) == 1:
            other = args[0]
            if isinstance(other, DictList):
                list.extend(self, other)
                self._dict = other._dict.copy()
            else:
                self.extend(other)

    def has_id(self, id):
        return id in self._dict

    def _check(self, id):
        """make sure duplicate id's are not added.
        This function is called before adding in elements.

        """
        if id in self._dict:
            raise ValueError("id %s is already present in list" % str(id))

    def _generate_index(self):
        """rebuild the _dict index"""
        self._dict = {v.id: k for k, v in enumerate(self)}

    def get_by_id(self, id):
        """return the element with a matching id"""
        return list.__getitem__(self, self._dict[id])

    def list_attr(self, attribute):
        """return a list of the given attribute for every object

        """
        return [getattr(i, attribute) for i in self]

    def query(self, search_function, attribute="id"):
        """query the list

        search_function: used to select which objects to return
            * a string, in which case any object.attribute containing
              the string will be returned

            * a compiled regular expression

            * a function which takes one argument and returns True
              for desired values

        attribute: the attribute to be searched for (default is 'id').
                   If this is None, the object itself is used.

        returns: a list of objects which match the query
        """
        if attribute is None:
            def select_attribute(x):
                return x
        else:
            def select_attribute(x):
                return getattr(x, attribute)

        # if the search_function is a regular expression
        if isinstance(search_function, str):
            search_function = re.compile(search_function)
        if hasattr(search_function, "findall"):
            matches = (i for i in self
                       if search_function.findall(select_attribute(i)) != [])
        else:
            matches = (i for i in self
                       if search_function(select_attribute(i)))
        results = self.__class__()
        results._extend_nocheck(matches)
        return results

    def _replace_on_id(self, new_object):
        """Replace an object by another with the same id."""
        the_id = new_object.id
        the_index = self._dict[the_id]
        list.__setitem__(self, the_index, new_object)

    # overriding default list functions with new ones
    def append(self, object):
        """append object to end"""
        the_id = object.id
        self._check(the_id)
        self._dict[the_id] = len(self)
        list.append(self, object)

    def union(self, iterable):
        """adds elements with id's not already in the model"""
        _dict = self._dict
        append = self.append
        for i in iterable:
            if i.id not in _dict:
                append(i)

    def extend(self, iterable):
        """extend list by appending elements from the iterable"""
        # Sometimes during initialization from an older pickle, _dict
        # will not have initialized yet, because the initialization class was
        # left unspecified. This is an issue because unpickling calls
        # DictList.extend, which requires the presence of _dict. Therefore,
        # the issue is caught and addressed here.
        if not hasattr(self, "_dict") or self._dict is None:
            self._dict = {}
        _dict = self._dict
        current_length = len(self)
        list.extend(self, iterable)
        for i, obj in enumerate(islice(self, current_length, None),
                                current_length):
            the_id = obj.id
            if the_id not in _dict:
                _dict[the_id] = i
            else:
                # undo the extend and raise an error
                self = self[:current_length]
                self._check(the_id)
                # if the above succeeded, then the id must be present
                # twice in the list being added
                raise ValueError("id '%s' at index %d is non-unique. "
                                 "Is it present twice?" % (str(the_id), i))

    def _extend_nocheck(self, iterable):
        """extends without checking for uniqueness

        This function should only be used internally by DictList when it
        can guarentee elements are already unique (as in when coming from
        self or other DictList). It will be faster because it skips these
        checks.

        """
        current_length = len(self)
        list.extend(self, iterable)
        _dict = self._dict
        if current_length is 0:
            self._generate_index()
            return
        for i, obj in enumerate(islice(self, current_length, None),
                                current_length):
            _dict[obj.id] = i

    def __add__(self, other):
        """x.__add__(y) <==> x + y

        other: iterable
            other must contain only unique id's which do not intersect
            with self

        """
        total = DictList()
        total.extend(self)
        total.extend(other)
        return total

    def __iadd__(self, other):
        """x.__iadd__(y) <==> x += y

        other: iterable
            other must contain only unique id's whcih do not intersect
            with self

        """
        self.extend(other)
        return self

    def __reduce__(self):
        return (self.__class__, (), self.__getstate__(), self.__iter__())

    def __getstate__(self):
        """gets internal state

        This is only provided for backwards compatibilty so older
        versions of cobrapy can load pickles generated with cobrapy. In
        reality, the "_dict" state is ignored when loading a pickle"""
        return {"_dict": self._dict}

    def __setstate__(self, state):
        """sets internal state

        Ignore the passed in state and recalculate it. This is only for
        compatibility with older pickles which did not correctly specify
        the initialization class"""
        self._generate_index()

    def index(self, id, *args):
        """Determine the position in the list

        id: A string or a :class:`~cobra.core.Object.Object`

        """
        # because values are unique, start and stop are not relevant
        if isinstance(id, string_types):
            try:
                return self._dict[id]
            except KeyError:
                raise ValueError("%s not found" % id)
        try:
            i = self._dict[id.id]
            if self[i] is not id:
                raise ValueError(
                    "Another object with the identical id (%s) found" % id.id)
            return i
        except KeyError:
            raise ValueError("%s not found" % str(id))

    def __contains__(self, object):
        """DictList.__contains__(object) <==> object in DictList

        object: str or :class:`~cobra.core.Object.Object`

        """
        if hasattr(object, "id"):
            the_id = object.id
        # allow to check with the object itself in addition to the id
        else:
            the_id = object
        return the_id in self._dict

    def __copy__(self):
        the_copy = DictList()
        list.extend(the_copy, self)
        the_copy._dict = self._dict.copy()
        return the_copy

    def insert(self, index, object):
        """insert object before index"""
        self._check(object.id)
        list.insert(self, index, object)
        # all subsequent entries now have been shifted up by 1
        _dict = self._dict
        for i, j in iteritems(_dict):
            if j >= index:
                _dict[i] = j + 1
        _dict[object.id] = index

    def pop(self, *args):
        """remove and return item at index (default last)."""
        value = list.pop(self, *args)
        index = self._dict.pop(value.id)
        # If the pop occured from a location other than the end of the list,
        # we will need to subtract 1 from every entry afterwards
        if len(args) == 0 or args == [-1]:  # removing from the end of the list
            return value
        _dict = self._dict
        for i, j in iteritems(_dict):
            if j > index:
                _dict[i] = j - 1
        return value

    def remove(self, x):
        """.. warning :: Internal use only"""
        # Each item is unique in the list which allows this
        # It is much faster to do a dict lookup than n string comparisons
        self.pop(self.index(x))

    # these functions are slower because they rebuild the _dict every time
    def reverse(self):
        """reverse *IN PLACE*"""
        list.reverse(self)
        self._generate_index()

    def sort(self, cmp=None, key=None, reverse=False):
        """stable sort *IN PLACE*

        cmp(x, y) -> -1, 0, 1

        """
        if key is None:
            def key(i):
                return i.id
        if PY3:
            list.sort(self, key=key, reverse=reverse)
        else:
            list.sort(self, cmp=cmp, key=key, reverse=reverse)
        self._generate_index()

    def __getitem__(self, i):
        if isinstance(i, int):
            return list.__getitem__(self, i)
        elif isinstance(i, slice):
            selection = self.__class__()
            selection._extend_nocheck(list.__getitem__(self, i))
            return selection
        elif hasattr(i, "__len__"):
            if len(i) == len(self) and isinstance(i[0], (bool, bool_)):
                selection = self.__class__()
                result = (o for j, o in enumerate(self) if i[j])
                selection._extend_nocheck(result)
                return selection
            else:
                return self.__class__(list.__getitem__(self, i))
        else:
            return list.__getitem__(self, i)

    def __setitem__(self, i, y):
        if isinstance(i, slice):
            # In this case, y needs to be a list. We will ensure all
            # the id's are unique
            for obj in y:  # need to be setting to a list
                self._check(obj.id)
                # Insert a temporary placeholder so we catch the presence
                # of a duplicate in the items being added
                self._dict[obj.id] = None
            list.__setitem__(self, i, y)
            self._generate_index()
            return
        self._dict.pop(self[i].id)
        the_id = y.id
        self._check(the_id)
        list.__setitem__(self, i, y)
        self._dict[the_id] = i

    def __delitem__(self, index):
        removed = self[index]
        list.__delitem__(self, index)
        if isinstance(removed, list):
            self._generate_index()
            return
        _dict = self._dict
        _dict.pop(removed.id)
        for i, j in iteritems(_dict):
            if j > index:
                _dict[i] = j - 1

    def __getslice__(self, i, j):
        return self.__getitem__(slice(i, j))

    def __setslice__(self, i, j, y):
        self.__setitem__(slice(i, j), y)

    def __delslice__(self, i, j):
        self.__delitem__(slice(i, j))

    def __getattr__(self, attr):
        try:
            return DictList.get_by_id(self, attr)
        except KeyError:
            raise AttributeError("DictList has no attribute or entry %s" %
                                 (attr))

    def __dir__(self):
        # override this to allow tab complete of items by their id
        attributes = dir(self.__class__)
        attributes.append("_dict")
        attributes.extend(self._dict.keys())
        return attributes



# Solvers are expected to follow the following interface
# create_problem: makes a solver problem object from a cobra.model and
# sets parameters (if possible)

# format_solution: Returns a cobra.Solution object.  This is where one
# should dress the cobra.model with results if desired.

# get_status: converts a solver specific status flag to a cobra pie flag.

# set_parameter: takes solver specific parameter strings and sets them.

# solve: solves the optimization problem.  this is where one should put
# in logic on what to try if the problem
# isn't optimal

# solve_problem: dumb and fast which will set parameters, if provided

# update_problem: changes bounds and linear objective coefficient of the
# solver specific problem file, given the complementary cobra.model

# This attempts to import all working solvers in this directory

from __future__ import absolute_import
from warnings import warn
from os import listdir, path

solver_dict = {}
possible_solvers = set()


def add_solver(solver_name, use_name=None):
    """add a solver module to the solvers"""
    exec("from . import " + solver_name)
    solver = eval(solver_name)
    if use_name is None:
        if hasattr(solver, "solver_name"):
            use_name = solver.solver_name
        else:
            use_name = solver_name
    solver_dict[use_name] = eval(solver_name)

for i in listdir(path.dirname(path.abspath(__file__))):
    if i.startswith("_") or i.startswith(".") or i.startswith('legacy'):
        continue
    if i.startswith("parameters"):
        continue
    if i.endswith(".py") or i.endswith(".so") or i.endswith(".pyc") \
            or i.endswith(".pyd"):
        possible_solvers.add(i.split(".")[0])

if "wrappers" in possible_solvers:
    possible_solvers.remove("wrappers")

for solver in possible_solvers:
    try:
        add_solver(solver)
    except:
        pass
    del solver

if len(solver_dict) == 0:
    warn("No LP solvers found")

# clean up the namespace
del path, listdir, warn, i, possible_solvers


class SolverNotFound(Exception):
    None


def get_solver_name(mip=False, qp=False):
    """returns a solver name

    raises SolverNotFound if a suitable solver is not found
    """
    if len(solver_dict) == 0:
        raise SolverNotFound("no solvers installed")
    # glpk only does lp, not qp. Gurobi and cplex are better at mip
    mip_order = ["gurobi", "cplex", "mosek", "coin", "cglpk", "glpk"]
    lp_order = ["cglpk", "cplex",  "gurobi", "mosek", "coin", "glpk"]
    qp_order = ["gurobi", "cplex", "mosek"]

    if mip is False and qp is False:
        for solver_name in lp_order:
            if solver_name in solver_dict:
                return solver_name
        # none of them are in the list order - so return the first one
        return list(solver_dict)[0]
    elif qp:  # mip does not yet matter for this determination
        for solver_name in qp_order:
            if solver_name in solver_dict:
                return solver_name
        # see if any solver defines set_quadratic_objective
        for solver_name in solver_dict:
            if hasattr(solver_dict[solver_name], "set_quadratic_objective"):
                return solver_name
        raise SolverNotFound("no qp-capable solver found")
    else:
        for solver_name in mip_order:
            if solver_name in solver_dict:
                return solver_name
        for solver_name in solver_dict:
            if hasattr(solver_dict[solver_name], "_SUPPORTS_MIP"):
                return solver_name
    raise SolverNotFound("no mip-capable solver found")


def optimize(cobra_model, solver=None, **kwargs):
    """Wrapper to optimization solvers

    solver : str
        Name of the LP solver from solver_dict to use. If None is given, the
        default one will be used

    """
    # If the default solver is not installed then use one of the others
    if solver is None:
        qp = "quadratic_component" in kwargs and \
            kwargs["quadratic_component"] is not None
        solver = get_solver_name(qp=qp)

    return solver_dict[solver].solve(cobra_model, **kwargs)



# PLEASE NOTE THAT JYTHON SUPPORT (and this jython-only-solver) is deprecated
#Interface to ilog/cplex 12.4 python / jython interfaces
#QPs are not yet supported under jython
from __future__ import print_function
from os import name as __name
from copy import deepcopy
from warnings import warn
###solver specific parameters
from .parameters import status_dict, variable_kind_dict, \
     sense_dict, parameter_mappings, parameter_defaults, \
     objective_senses, default_objective_sense

from ..core.Solution import Solution
from time import time
from six import iteritems
solver_name = 'cplex'
parameter_defaults = parameter_defaults[solver_name]
sense_dict = eval(sense_dict[solver_name])

from ilog.cplex import IloCplex
from ilog.cplex.IloCplex import DoubleParam, IntParam, StringParam
from ilog.concert import IloNumVarType, IloObjectiveSense 
#__solver_class = IloCplex
status_dict = eval(status_dict[solver_name])
class Problem(IloCplex):
    def __init__(self):
        IloCplex.__init__(self)
        self._lp_matrix = self.addLPMatrix()
        self.objective_value = None
        self._objective_sense = 'maximize'
    def add_linear_expression(self, linear_expression, metabolite):
        b = metabolite._bound
        c = metabolite._constraint_sense
        the_id = metabolite.id
        if c == 'E':
            p = self.addEq(linear_expression, b, the_id)
        elif c == 'L':
            p = self.addLe(linear_expression, b, the_id)
        elif c == 'G':
            p = self.addGe(linear_expression, b, the_id)
        else:
            raise Exception("Constraint sense '%s' for metabolite %s is not valid"%(c,
                                                                                    the_id))
        return(p)

__solver_class = Problem
parameter_mappings = parameter_mappings['%s_%s'%(solver_name,
                                                 __name)]
variable_kind_dict = eval(variable_kind_dict['%s_%s'%(solver_name,
                                                      __name)])
objective_senses = objective_senses['%s_%s'%(solver_name,
                                             __name)]
## from jarray import array as j_array
## def array(x, variable_type='d'):
##     return j_array(x, variable_type)


def get_status(lp):
    status = repr(lp.status).lower()
    if status in status_dict:
        status = status_dict[status]
    else:
        status = 'failed'
    return status

def set_parameter(lp, parameter_name, parameter_value):
    if parameter_name == 'objective_sense':
        objective = lp.getObjective()
        if objective is not None:
            objective.setSense(eval(objective_senses[parameter_value]))

    else:
        if hasattr(DoubleParam, parameter_name):
            parameter_type = 'DoubleParam'
        elif hasattr(IntParam, parameter_name):
            parameter_type = 'IntParam'
        elif hasattr(StringParam, parameter_name):
            parameter_type = 'StringParam'
        else:
            raise Exception("%s is not a DoubleParam, IntParam, or StringParam"%parameter_name)
        lp.setParam(eval('%s.%s'%(parameter_type, parameter_name)),
                    parameter_value)

def get_objective_value(lp):
    return lp.getObjValue()

def format_solution(lp, cobra_model, **kwargs):
    """
    TODO
    """
    status = get_status(lp)
    try:
        x = lp.getValues(lp.variables)
        x_dict = dict(zip(cobra_model.reactions, x))
        objective_value = lp.getObjValue()
    except:
        x = x_dict = objective_value = None
        #print status

    try:
        y = lp.getDuals(lp.variables)
        y_dict = dict(zip(cobra_model.metabolites, y))
    except:
        y = y_dict = None
    return(Solution(objective_value, x=x, x_dict=x_dict, y=y,
                    y_dict=y_dict, status=status))

def create_problem(cobra_model,  **kwargs):
    """Solver-specific method for constructing a solver problem from
    a cobra.Model.  This can be tuned for performance using kwargs

    TODO: This will need to be specific for python / jython
    """
    the_parameters = parameter_defaults
    if kwargs:
        the_parameters = deepcopy(parameter_defaults)
        the_parameters.update(kwargs)

    lp = Problem()

    if 'log_file' not in the_parameters:
        lp.setWarning(None)
        lp.setOut(None)
    [set_parameter(lp, parameter_mappings[k], v)
     for k, v in iteritems(the_parameters) if k in parameter_mappings]
    quadratic_component = the_parameters['quadratic_component']
    new_objective = the_parameters['new_objective']
    error_reporting = the_parameters['error_reporting']
    lp._objective_sense  = the_parameters['objective_sense']
    if 'relax_b' in the_parameters:
        warn('need to reimplement relax_b')
        relax_b = False
    else:
        relax_b = False

    #Using the new objects
    #NOTE: This might be slow
    objective_coefficients = []
    lower_bounds = []
    upper_bounds = []
    variable_names = []
    variable_kinds = []
    [(objective_coefficients.append(x.objective_coefficient),
      lower_bounds.append(x.lower_bound),
      upper_bounds.append(x.upper_bound),
      variable_names.append(x.id),
      variable_kinds.append(variable_kind_dict[x.variable_kind]))
     for x in cobra_model.reactions]

    #Only add the variable types if one or more variables is an integer, just
    #in case the java interface has the same bug as the python interface where
    #the problem type switches to integer if variable types are supplied even
    #if all are continuous
    if variable_kind_dict['integer'] in variable_kinds:
        lp.variables = lp.numVarArray(len(cobra_model.reactions), lower_bounds,
                                      upper_bounds, variable_kinds, variable_names)
    else:
        lp.variables = lp.numVarArray(len(cobra_model.reactions), lower_bounds, upper_bounds,
                                      variable_names)
    
    lp.variable_dict = dict(zip(cobra_model.reactions, lp.variables))
    if lp._objective_sense == 'maximize':
        __lp_add_objective = lp.addMaximize
    else:
        __lp_add_objective = lp.addMinimize

    __lp_add_objective(lp.scalProd(lp.variables, objective_coefficients))
    

    
    lp.constraints = []
    lp.constraint_dict = {}
    for the_metabolite in cobra_model.metabolites:
        linear_expression = lp.sum([lp.prod(k._metabolites[the_metabolite],
                                            lp.variable_dict[k])
                                    for k in the_metabolite._reaction])
        expression_pointer = lp.add_linear_expression(linear_expression, the_metabolite)
        lp.constraints.append(expression_pointer)
        lp.constraint_dict[the_metabolite] = expression_pointer
    
    if quadratic_component is not None:
        raise Exception("cplex through java isn't configured for QPs yet")
        if not hasattr(quadratic_component, 'todok'):
            raise Exception('quadratic component must have method todok')
        quadratic_component_scaled = quadratic_component.todok()

        lp.parameters.emphasis.numerical.set(1)
        for k, v in quadratic_component_scaled.items():
            lp.objective.set_quadratic_coefficients(int(k[0]), int(k[1]), v)

    ## #Set the problem type as cplex doesn't appear to do this correctly
    ## problem_type = Cplex.problem_type.LP
    ## if Cplex.variables.type.integer in variable_kinds:
    ##     if quadratic_component is not None:
    ##         problem_type = Cplex.problem_type.MIQP
    ##     else:
    ##         problem_type = Cplex.problem_type.MILP
    ## elif quadratic_component is not None:
    ##     problem_type = Cplex.problem_type.QP
    ## lp.set_problem_type(problem_type)
    return(lp)


def update_problem(lp, cobra_model, **kwargs):
    """A performance tunable method for updating a model problem file

    lp: A gurobi problem object

    cobra_model: the cobra.Model corresponding to 'lp'

    """
    #When reusing the basis only assume that the objective coefficients or bounds can change
    #BUG with changing / unchanging the basis
    try:
        new_objective = kwargs['new_objective']
    except:
        new_objective = None
    try:
        update_problem_reaction_bounds = kwargs['update_problem_reaction_bounds']
    except:
        update_problem_reaction_bounds = True
    try:
        quadratic_component = kwargs['quadratic_component']
        if quadratic_component is not None:
            warn("update_problem does not yet take quadratic_component as a parameter")
    except:
        quadratic_component = None

    the_objective = lp.getObjective()
    for the_variable, the_reaction in zip(lp.variables, cobra_model.reactions):
        the_variable.setUB(float(the_reaction.upper_bound))
        the_variable.setLB(float(the_reaction.lower_bound))
        the_objective.setLinearCoef(the_variable, the_reaction.objective_coefficient)
            


###
def solve_problem(lp, **kwargs):
    """A performance tunable method for solving a problem

    """
    #Update parameter settings if provided
    if kwargs:
        [set_parameter(lp, parameter_mappings[k], v)
         for k, v in iteritems(kwargs) if k in parameter_mappings]
    try:
        the_problem = kwargs['the_problem']
    except:
        the_problem = False
    if isinstance(the_problem, __solver_class):
        try:
            the_basis = the_problem.solution.basis.get_basis()
            lp.start.set_basis(the_basis[0],the_basis[1])
            lp.parameters.preprocessing.presolve.set(0)
        except:
            warn("cplex_java isn't yet configured to reuse the basis")
    
    lp.solve()
    #If the solver takes more than 0.1 s with a hot start it is likely stuck
    status = get_status(lp)
    return status

    
def solve(cobra_model, **kwargs):
    """

    """
    #Start out with default parameters and then modify if
    #new onese are provided
    the_parameters = deepcopy(parameter_defaults)
    if kwargs:
        the_parameters.update(kwargs)
    #Update objectives if they are new.
    if 'new_objective' in the_parameters:
        raise ValueError("new_objective option removed")

    if 'the_problem' in the_parameters:
        the_problem = the_parameters['the_problem']
    else:
        the_problem = None
    if 'error_reporting' in the_parameters:
        error_reporting = the_parameters['error_reporting']
    else:
        error_reporting = False
    if isinstance(the_problem, __solver_class):
        #Update the problem with the current cobra_model
        lp = the_problem
        update_problem(lp, cobra_model, **the_parameters)
    else:
        #Create a new problem
        lp = create_problem(cobra_model, **the_parameters)
    #Deprecated way for returning a solver problem created from a cobra_model
    #without performing optimization
    if the_problem == 'setup':
        return lp

    ###Try to solve the problem using other methods if the first method doesn't work
    try:
        lp_method = the_parameters['lp_method']
    except:
        lp_method = 1
    the_methods = [1, 2, 3, 4, 5, 6]
    if lp_method in the_methods:
        the_methods.remove(lp_method)
    #Start with the user specified method
    the_methods.insert(0, lp_method)
    for the_method in the_methods:
        the_parameters['lp_method'] = the_method
        try:
            status = solve_problem(lp, **the_parameters)
        except:
            status = 'failed'
        if status == 'optimal':
            break

    the_solution = format_solution(lp, cobra_model)
    if status != 'optimal' and error_reporting:
        print('{:s} failed: {:s}'.format(solver_name, status))
    cobra_model.solution = the_solution
    solution = {'the_problem': lp, 'the_solution': the_solution}
    return solution



from unittest import TestCase, TestLoader, TextTestRunner, skipIf

import sys

if __name__ == "__main__":
    sys.path.insert(0, "../..")
    from cobra.test import create_test_model, data_directory
    from cobra.design import *
    from cobra.design.design_algorithms import _add_decision_variable
    from cobra.solvers import get_solver_name
    sys.path.pop(0)
else:
    from . import create_test_model, data_directory
    from ..design import *
    from ..design.design_algorithms import _add_decision_variable
    from ..solvers import get_solver_name

try:
    solver = get_solver_name(mip=True)
except:
    no_mip_solver = True
else:
    no_mip_solver = False


class TestDesignAlgorithms(TestCase):
    """Test functions in cobra.design"""

    def test_dual(self):
        model = create_test_model("textbook")
        self.assertAlmostEqual(model.optimize("maximize").f, 0.874, places=3)
        dual = dual_problem(model)
        self.assertAlmostEqual(dual.optimize("minimize").f, 0.874, places=3)

    def test_dual_integer_vars_as_lp(self):
        model = create_test_model("textbook")
        var = _add_decision_variable(model, "AKGDH")
        self.assertAlmostEqual(model.optimize("maximize").f, 0.874, places=3)
        # as lp: make integer continuous, set to 1
        dual = dual_problem(model, "maximize", [var.id], copy=True)
        r = dual.reactions.get_by_id(var.id)
        r.variable_kind = "continuous"
        r.lower_bound = r.upper_bound = 1
        self.assertAlmostEqual(dual.optimize("minimize").f, 0.874, places=3)
        r.lower_bound = r.upper_bound = 0
        self.assertAlmostEqual(dual.optimize("minimize").f, 0.858, places=3)

    @skipIf(no_mip_solver, "no MILP solver found")
    def test_dual_integer_vars_as_mip(self):
        # mip
        model = create_test_model("textbook")
        var = _add_decision_variable(model, "AKGDH")
        dual = dual_problem(model, "maximize", [var.id], copy=True)
        var_in_dual = dual.reactions.get_by_id(var.id)

        # minimization, so the optimal value state is to turn off AKGDH
        self.assertAlmostEqual(dual.optimize("minimize").f, 0.858, places=3)

        # turn off AKGDH in dual
        var_in_dual.lower_bound = var_in_dual.upper_bound = 1
        self.assertAlmostEqual(dual.optimize("minimize").f, 0.874, places=3)

        # turn on AKGDH in dual
        var_in_dual.lower_bound = var_in_dual.upper_bound = 0
        self.assertAlmostEqual(dual.optimize("minimize").f, 0.858, places=3)

    @skipIf(no_mip_solver, "no MILP solver found")
    def test_optknock(self):
        model = create_test_model("textbook")
        model.reactions.get_by_id("EX_o2_e").lower_bound = 0
        knockable_reactions = ["ACKr", "AKGDH", "ACALD", "LDH_D"]
        optknock_problem = set_up_optknock(model, "EX_lac__D_e",
                                           knockable_reactions, n_knockouts=2,
                                           copy=False)
        solution = run_optknock(optknock_problem, tolerance_integer=1e-9)
        self.assertIn("ACKr", solution.knockouts)
        self.assertIn("ACALD", solution.knockouts)
        self.assertAlmostEqual(solution.f, 17.891, places=3)

# make a test suite to run all of the tests
loader = TestLoader()
suite = loader.loadTestsFromModule(sys.modules[__name__])


def test_all():
    TextTestRunner(verbosity=2).run(suite)

if __name__ == "__main__":
    test_all()
